# Copyright (C) 2008-2010 Adam Olsen
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
#
# The developers of the Exaile media player hereby grant permission
# for non-GPL compatible GStreamer and Exaile plugins to be used and
# distributed together with GStreamer and Exaile. This permission is
# above and beyond the permissions granted by the GPL license by which
# Exaile is covered. If you modify this code, you may extend this
# exception to your version of the code, but you are not obligated to
# do so. If you do not wish to do so, delete this exception statement
# from your version.

from gi.repository import Gdk
from gi.repository import GdkPixbuf
from gi.repository import GLib
from gi.repository import GObject
from gi.repository import Gtk
import itertools
import logging

from xl.nls import gettext as _
from xl import common, event, formatter, settings, trax
import xlgui
from xlgui import guiutil, icons, panel
from xlgui.panel import menus
from xlgui.widgets import menu
from xlgui.widgets.common import DragTreeView

logger = logging.getLogger(__name__)

# TODO: come up with a more customizable way to handle this
SEARCH_TAGS = ("artist", "albumartist", "album", "title")


def first_meaningful_char(s):
    # Keep explicit str() conversion in case we ever end up receiving
    # a non-string sort tag (e.g. an int for track number)
    for c in str(s):
        if c.isdigit():
            return '0'
        elif c.isalpha():
            return c
    else:
        return '_'


class Order:
    """
    An Order represents a structure for arranging Tracks into the
    Collection tree.

    It is based on a list of levels, which each take the form (("sort1",
    "sort2"), "$displaytag - $displaytag", ("search1", "search2")) wherin
    the first entry is a tuple of tags to use for sorting, the second a
    format string for xl.formatter, and the third a tuple of tags to use
    for searching.

    When passed in the parameters, a level can also be a single string
    instead of a tuple, and it will be treated equivalently to (("foo",),
    "$foo", ("foo",)) for some string "foo".
    """

    def __init__(self, name, levels, use_compilations=True):
        self.__name = name
        self.__levels = [self.__parse_level(l) for l in levels]
        self.__formatters = [formatter.TrackFormatter(l[1]) for l in self.__levels]
        self.__use_compilations = use_compilations

    @staticmethod
    def __parse_level(val):
        if isinstance(val, str):
            val = ((val,), "$%s" % val, (val,))
        return tuple(val)

    @property
    def name(self):
        return self.__name

    @property
    def use_compilations(self):
        return self.__use_compilations

    def get_levels(self):
        return self.__levels[:]

    def __len__(self):
        return len(self.__levels)

    def __eq__(self, other):
        return self.__levels == other.get_levels()

    def all_sort_tags(self):
        return set(itertools.chain(*[l[0] for l in self.__levels]))

    def get_sort_tags(self, level):
        return list(self.__levels[level][0])

    def all_search_tags(self):
        return list(itertools.chain(*[l[2] for l in self.__levels]))

    def get_search_tags(self, level):
        return list(set(self.__levels[level][2]))

    def format_track(self, level, track):
        return self.__formatters[level].format(track)


DEFAULT_ORDERS = [
    # fmt: off
    Order(_("Artist"),
          ("artist", "album",
           (("discnumber", "tracknumber", "title"), "$title", ("title",)))),
    Order(_("Album Artist"),
          ("albumartist", "album",
           (("discnumber", "tracknumber", "title"), "$title", ("title",)))),
    Order(_("Album"),
          ("album",
           (("discnumber", "tracknumber", "title"), "$title", ("title",)))),
    Order(_("Genre - Artist"),
          ('genre', 'artist', 'album',
           (("discnumber", "tracknumber", "title"), "$title", ("title",)))),
    Order(_("Genre - Album Artist"),
          ('genre', 'albumartist', 'album',
           (("discnumber", "tracknumber", "title"), "$title", ("title",)))),
    Order(_("Genre - Album"),
          ('genre', 'album', 'artist',
           (("discnumber", "tracknumber", "title"), "$title", ("title",)))),
    Order(_("Date - Artist"),
          ('date', 'artist', 'album',
           (("discnumber", "tracknumber", "title"), "$title", ("title",)))),
    Order(_("Date - Album Artist"),
          ('date', 'albumartist', 'album',
           (("discnumber", "tracknumber", "title"), "$title", ("title",)))),
    Order(_("Date - Album"),
          ('date', 'album', 'artist',
           (("discnumber", "tracknumber", "title"), "$title", ("title",)))),
    Order(_("Artist - (Date - Album)"),
          ('artist',
           (('date', 'album'), "$date - $album", ('date', 'album')),
           (("discnumber", "tracknumber", "title"), "$title", ("title",)))),
    Order(_("Album Artist - (Date - Album)"),
          ('albumartist',
           (('date', 'album'), "$date - $album", ('date', 'album')),
           (("discnumber", "tracknumber", "title"), "$title", ("title",)))),
    # fmt: on
]


class CollectionPanel(panel.Panel):
    """
    The collection panel
    """

    __gsignals__ = {
        'append-items': (GObject.SignalFlags.RUN_LAST, None, (object, bool)),
        'replace-items': (GObject.SignalFlags.RUN_LAST, None, (object,)),
        'queue-items': (GObject.SignalFlags.RUN_LAST, None, (object,)),
        'collection-tree-loaded': (GObject.SignalFlags.RUN_LAST, None, ()),
    }

    ui_info = ('collection.ui', 'CollectionPanel')

    def __init__(
        self,
        parent,
        collection,
        name=None,
        _show_collection_empty_message=False,
        label=_('Collection'),
    ):
        """
        Initializes the collection panel

        @param parent: the parent dialog
        @param collection: the xl.collection.Collection instance
        @param name: an optional name for this panel
        """
        panel.Panel.__init__(self, parent, name, label)

        self._show_collection_empty_message = _show_collection_empty_message
        self.collection = collection
        self.use_alphabet = settings.get_option('gui/use_alphabet', True)
        self.panel_stack = self.builder.get_object('CollectionPanel')
        self.panel_content = self.builder.get_object('CollectionPanelContent')
        self.panel_empty = self.builder.get_object('CollectionPanelEmpty')
        self.choice = self.builder.get_object('collection_combo_box')
        self._search_num = 0
        self._refresh_id = 0
        self.start_count = 0
        self.keyword = ''
        self.orders = DEFAULT_ORDERS[:]
        self._setup_tree()
        self._setup_widgets()
        self._check_collection_empty()
        self._setup_images()
        self._connect_events()
        self.order = None
        self.tracks = []
        self.sorted_tracks = []

        event.add_ui_callback(
            self._check_collection_empty, 'libraries_modified', collection
        )

        self.menu = menus.CollectionContextMenu(self)

        self.load_tree()

    def _setup_widgets(self):
        """
        Sets up the various widgets to be used in this panel
        """
        self.choice = self.builder.get_object('collection_combo_box')
        self.choicemodel = self.builder.get_object('collection_combo_model')
        self.repopulate_choices()

        self.filter = guiutil.SearchEntry(
            self.builder.get_object('collection_search_entry')
        )

    def repopulate_choices(self):
        self.choice.set_model(None)
        self.choicemodel.clear()
        for order in self.orders:
            self.choicemodel.append([order.name])
        self.choice.set_model(self.choicemodel)
        # FIXME: use something other than index here, since index
        # doesn't deal well with dynamic lists...
        active = settings.get_option('gui/collection_active_view', 0)
        if len(self.orders) - 1 < active:
            active = len(self.orders) - 1
        if active < 0:
            active = 0
        self.choice.set_active(active)

    def _check_collection_empty(self, *e):
        if self._show_collection_empty_message and not self.collection.libraries:
            # should show empty panel
            if self.panel_stack.get_visible_child() == self.panel_content:
                self.panel_stack.set_visible_child(self.panel_empty)
        else:
            # should show content panel
            if self.panel_stack.get_visible_child() == self.panel_empty:
                self.panel_stack.set_visible_child(self.panel_content)

    def _connect_events(self):
        """
        Uses signal_autoconnect to connect the various events
        """
        self.builder.connect_signals(
            {
                'on_collection_combo_box_changed': lambda *e: self.load_tree(),
                'on_refresh_button_press_event': self.on_refresh_button_press_event,
                'on_refresh_button_key_press_event': self.on_refresh_button_key_press_event,
                'on_collection_search_entry_activate': self.on_collection_search_entry_activate,
                'on_add_music_button_clicked': self.on_add_music_button_clicked,
            }
        )
        self.tree.connect('key-release-event', self.on_key_released)
        event.add_ui_callback(self.refresh_tags_in_tree, 'track_tags_changed')
        event.add_ui_callback(
            self.refresh_tracks_in_tree, 'tracks_added', self.collection
        )
        event.add_ui_callback(
            self.refresh_tracks_in_tree, 'tracks_removed', self.collection
        )

    def on_refresh_button_press_event(self, button, event):
        """
        Called on mouse activation of the refresh button
        """
        if event.triggers_context_menu():
            m = menu.Menu(None)
            m.attach_to_widget(button)
            m.add_simple(
                _('Rescan Collection'),
                xlgui.get_controller().on_rescan_collection,
                Gtk.STOCK_REFRESH,
            )
            m.popup(event)
            return

        if event.get_state() & Gdk.ModifierType.SHIFT_MASK:
            xlgui.get_controller().on_rescan_collection(None)
        else:
            self.load_tree()

    def on_refresh_button_key_press_event(self, widget, event):
        """
        Called on key presses on the refresh button
        """
        if event.keyval != Gdk.KEY_Return:
            return False

        if event.get_state() & Gdk.ModifierType.SHIFT_MASK:
            xlgui.get_controller().on_rescan_collection(None)
        else:
            self.load_tree()

    def on_key_released(self, widget, event):
        """
        Called when a key is released in the tree
        """
        if event.keyval == Gdk.KEY_Menu:
            Gtk.Menu.popup(self.menu, None, None, None, None, 0, event.time)
            return True

        if event.keyval == Gdk.KEY_Left:
            mods, paths = self.tree.get_selection().get_selected_rows()
            for path in paths:
                self.tree.collapse_row(path)
            return True

        if event.keyval == Gdk.KEY_Right:
            mods, paths = self.tree.get_selection().get_selected_rows()
            for path in paths:
                self.tree.expand_row(path, False)
            return True

        if event.keyval == Gdk.KEY_Return:
            self.append_to_playlist()
            return True
        return False

    def on_collection_search_entry_activate(self, entry):
        """
        Searches tracks and reloads the tree
        """
        self.keyword = entry.get_text()
        self.start_count += 1
        self.load_tree()

    def on_add_music_button_clicked(self, button):
        xlgui.get_controller().collection_manager()

    def _setup_images(self):
        """
        Sets up the various images that will be used in the tree
        """
        self.artist_image = icons.MANAGER.pixbuf_from_icon_name(
            'artist', Gtk.IconSize.SMALL_TOOLBAR
        )
        self.albumartist_image = icons.MANAGER.pixbuf_from_icon_name(
            'artist', Gtk.IconSize.SMALL_TOOLBAR
        )
        self.date_image = icons.MANAGER.pixbuf_from_icon_name(
            'office-calendar', Gtk.IconSize.SMALL_TOOLBAR
        )
        self.album_image = icons.MANAGER.pixbuf_from_icon_name(
            'image-x-generic', Gtk.IconSize.SMALL_TOOLBAR
        )
        self.title_image = icons.MANAGER.pixbuf_from_icon_name(
            'audio-x-generic', Gtk.IconSize.SMALL_TOOLBAR
        )
        self.genre_image = icons.MANAGER.pixbuf_from_icon_name(
            'genre', Gtk.IconSize.SMALL_TOOLBAR
        )

    def drag_data_received(self, *e):
        """
        stub
        """
        pass

    def drag_data_delete(self, *e):
        """
        stub
        """
        pass

    def drag_get_data(self, treeview, context, selection, target_id, etime):
        """
        Called when a drag source wants data for this drag operation
        """
        tracks = treeview.get_selected_tracks()

        for track in tracks:
            DragTreeView.dragged_data[track.get_loc_for_io()] = track

        uris = trax.util.get_uris_from_tracks(tracks)
        selection.set_uris(uris)

    def _setup_tree(self):
        """
        Sets up the tree widget
        """
        self.tree = CollectionDragTreeView(self)
        self.tree.set_headers_visible(False)
        scroll = Gtk.ScrolledWindow()
        scroll.set_policy(Gtk.PolicyType.AUTOMATIC, Gtk.PolicyType.AUTOMATIC)
        scroll.add(self.tree)
        scroll.set_shadow_type(Gtk.ShadowType.IN)
        self.panel_content.pack_start(scroll, True, True, 0)
        self.panel_content.show_all()

        selection = self.tree.get_selection()
        selection.set_mode(Gtk.SelectionMode.MULTIPLE)
        pb = Gtk.CellRendererPixbuf()
        cell = Gtk.CellRendererText()
        col = Gtk.TreeViewColumn('Text')
        col.pack_start(pb, False)
        col.pack_start(cell, True)
        col.set_attributes(pb, pixbuf=0)
        col.set_attributes(cell, text=1)
        self.tree.append_column(col)

        if settings.get_option('gui/ellipsize_text_in_panels', False):
            from gi.repository import Pango

            cell.set_property('ellipsize-set', True)
            cell.set_property('ellipsize', Pango.EllipsizeMode.END)

        self.tree.set_row_separator_func(
            (lambda m, i, d: m.get_value(i, 1) is None), None
        )

        self.model = Gtk.TreeStore(GdkPixbuf.Pixbuf, str, object)

        self.tree.connect("row-expanded", self.on_expanded)

    def _find_tracks(self, iter):
        """
        finds tracks matching a given iter.
        """
        self.load_subtree(iter)
        search = self.get_node_search_terms(iter)
        matcher = trax.TracksMatcher(search)
        srtrs = trax.search_tracks(self.tracks, [matcher])
        return [x.track for x in srtrs]

    def append_to_playlist(self, item=None, event=None, replace=False):
        """
        Adds items to the current playlist
        """
        if replace:
            self.emit('replace-items', self.tree.get_selected_tracks())
        else:
            self.emit('append-items', self.tree.get_selected_tracks(), True)

    def button_press(self, widget, event):
        """
        Called when the user clicks on the tree
        """
        # selection = self.tree.get_selection()
        x, y = [int(v) for v in event.get_coords()]
        # path = self.tree.get_path_at_pos(x, y)
        if event.type == Gdk.EventType._2BUTTON_PRESS:
            replace = settings.get_option('playlist/replace_content', False)
            self.append_to_playlist(replace=replace)
            return False
        elif event.button == Gdk.BUTTON_MIDDLE:
            self.append_to_playlist(replace=True)
            return False

    def on_expanded(self, tree, iter, path):
        """
        Called when a user expands a tree item.

        Loads the various nodes that belong under this node.
        """
        self.load_subtree(iter)

    def get_node_search_terms(self, node):
        """
        Finds all the related search terms for a particular node
        @param node: the node you wish to create search terms
        """
        if not node:
            return ""

        queries = []
        while node:
            queries.append(self.model.get_value(node, 2))
            node = self.model.iter_parent(node)

        return " ".join(queries)

    def refresh_tags_in_tree(self, type, track, tags):
        if (
            settings.get_option('gui/sync_on_tag_change', True)
            and bool(tags & self.order.all_sort_tags())
            and self.collection.loc_is_member(track.get_loc_for_io())
        ):
            self._refresh_tags_in_tree()

    def refresh_tracks_in_tree(self, type, obj, loc):
        self._refresh_tags_in_tree()

    @common.glib_wait(500)
    def _refresh_tags_in_tree(self):
        """
        Callback for when tags have changed and the tree
        needs reloading.
        """
        # Trying to reload while we're rescanning is really inefficient,
        # so we delay it until we're done scanning.
        if self.collection._scanning:
            return True
        self.resort_tracks()
        self.load_tree()
        return False

    def resort_tracks(self):
        # import time
        # print("sorting...", time.clock())
        self.sorted_tracks = trax.sort_tracks(
            self.order.get_sort_tags(0), self.collection.get_tracks()
        )
        # print("sorted.", time.clock())

    def load_tree(self):
        """
        Loads the Gtk.TreeView for this collection panel.

        Loads tracks based on the current keyword, or all the tracks in
        the collection associated with this panel
        """
        logger.debug("Reloading collection tree")
        self.current_start_count = self.start_count
        self.tree.set_model(None)
        self.model.clear()

        self.root = None
        oldorder = self.order
        self.order = self.orders[self.choice.get_active()]

        if not oldorder or oldorder != self.order:
            self.resort_tracks()

        # save the active view setting
        settings.set_option('gui/collection_active_view', self.choice.get_active())

        keyword = self.keyword.strip()
        tags = list(SEARCH_TAGS)
        tags += self.order.all_search_tags()
        tags = list(set(tags))  # uniquify list to speed up search

        self.tracks = list(
            trax.search_tracks_from_string(
                self.sorted_tracks, keyword, case_sensitive=False, keyword_tags=tags
            )
        )

        self.load_subtree(None)

        self.tree.set_model(self.model)

        self.emit('collection-tree-loaded')

    def _expand_node_by_name(self, search_num, parent, name, rest=None):
        """
        Recursive function to expand all nodes in a hierarchical list of
        names.

        @param search_num: the current search number
        @param parent: the parent node
        @param name: the name of the node to expand
        @param rest: the list of the nodes to expand after this one
        """
        iter = self.model.iter_children(parent)

        while iter:
            if search_num != self._search_num:
                return
            value = self.model.get_value(iter, 1)
            if not value:
                value = self.model.get_value(iter, 2)

            if value == name:
                self.tree.expand_row(self.model.get_path(iter), False)
                parent = iter
                break

            iter = self.model.iter_next(iter)

        if rest:
            item = rest.pop(0)
            GLib.idle_add(self._expand_node_by_name, search_num, parent, item, rest)

    def load_subtree(self, parent):
        """
        Loads all the sub nodes for a specified node

        @param node: the node
        """
        previously_loaded = False  # was the subtree already loaded
        iter_sep = None
        if parent is None:
            depth = 0
        else:
            if (
                self.model.iter_n_children(parent) != 1
                or self.model.get_value(self.model.iter_children(parent), 1) is not None
            ):
                previously_loaded = True
            iter_sep = self.model.iter_children(parent)
            depth = self.model.iter_depth(parent) + 1
        if previously_loaded:
            return

        search = self.get_node_search_terms(parent)

        try:
            tags = self.order.get_sort_tags(depth)
            matchers = [trax.TracksMatcher(search)]
            srtrs = trax.search_tracks(self.tracks, matchers)
            # sort only if we are not on top level, because tracks are
            # already sorted by fist order
            if depth > 0:
                srtrs = trax.sort_result_tracks(tags, srtrs)
        except IndexError:
            return  # at the bottom of the tree
        try:
            image = getattr(self, "%s_image" % tags[-1])
        except Exception:
            image = None
        bottom = False
        if depth == len(self.order) - 1:
            bottom = True

        display_counts = settings.get_option('gui/display_track_counts', True)
        draw_seps = settings.get_option('gui/draw_separators', True)
        last_char = ''
        last_val = ''
        last_dval = ''
        last_matchq = ''
        count = 0
        first = True
        path = None
        expanded = False
        to_expand = []

        for srtr in srtrs:
            # The value returned by get_tag_sort() may be of other
            # typa than str (e.g., an int for track number), hence
            # explicit conversion via str() is necessary.
            stagvals = [str(srtr.track.get_tag_sort(x)) for x in tags]
            stagval = " ".join(stagvals)
            if last_val != stagval or bottom:
                tagval = self.order.format_track(depth, srtr.track)
                match_query = " ".join(
                    [srtr.track.get_tag_search(t, format=True) for t in tags]
                )
                if bottom:
                    match_query += " " + srtr.track.get_tag_search("__loc", format=True)

                # Different *sort tags can cause stagval to not match
                # but the below code will produce identical entries in
                # the displayed tree.  This condition checks to ensure
                # that new entries are added if and only if they will
                # display different results, avoiding that problem.
                if match_query != last_matchq or tagval != last_dval or bottom:
                    if display_counts and path and not bottom:
                        iter = self.model.get_iter(path)
                        val = self.model.get_value(iter, 1)
                        val = "%s (%s)" % (val, count)
                        self.model.set_value(iter, 1, val)
                        count = 0

                    last_val = stagval
                    last_dval = tagval
                    if depth == 0 and draw_seps:
                        val = srtr.track.get_tag_sort(tags[0])
                        char = first_meaningful_char(val)
                        if first:
                            last_char = char
                        else:
                            if char != last_char and last_char != '':
                                self.model.append(parent, [None, None, None])
                            last_char = char
                    first = False

                    last_matchq = match_query
                    iter = self.model.append(parent, [image, tagval, match_query])
                    path = self.model.get_path(iter)
                    expanded = False
                    if not bottom:
                        self.model.append(iter, [None, None, None])
            count += 1
            if not expanded:
                alltags = []
                for i in range(depth + 1, len(self.order)):
                    alltags.extend(self.order.get_sort_tags(i))
                for t in alltags:
                    if t in srtr.on_tags:
                        # keep original path intact for following block
                        newpath = path
                        if depth > 0:
                            # for some reason, nested iters are always
                            # off by one in the terminal entry.
                            newpath = Gtk.TreePath.new_from_indices(
                                newpath[:-1] + [newpath[-1] - 1]
                            )
                        to_expand.append(newpath)
                        expanded = True

        if display_counts and path and not bottom:
            iter = self.model.get_iter(path)
            val = self.model.get_value(iter, 1)
            val = "%s (%s)" % (val, count)
            self.model.set_value(iter, 1, val)
            count = 0

        if (
            settings.get_option("gui/expand_enabled", True)
            and len(to_expand) < settings.get_option("gui/expand_maximum_results", 100)
            and len(self.keyword.strip())
            >= settings.get_option("gui/expand_minimum_term_length", 2)
        ):
            for row in to_expand:
                GLib.idle_add(self.tree.expand_row, row, False)

        if iter_sep is not None:
            self.model.remove(iter_sep)


class CollectionDragTreeView(DragTreeView):
    """
    Custom DragTreeView to retrieve data
    from collection tracks
    """

    def __init__(self, container, receive=False, source=True):
        """
        :param container: The container to place the TreeView into
        :param receive: True if the TreeView should receive drag events
        :param source: True if the TreeView should send drag events
        """
        DragTreeView.__init__(self, container, receive, source)

        self.set_has_tooltip(True)
        self.connect('query-tooltip', self.on_query_tooltip)

    def get_selection_empty(self):
        '''Returns True if there are no selected items'''
        return self.get_selection().count_selected_rows() == 0

    def get_selected_tracks(self):
        """
        Returns the currently selected tracks
        """
        model, paths = self.get_selection().get_selected_rows()

        if len(paths) == 0:
            return []

        tracks = set()
        for path in paths:
            iter = model.get_iter(path)
            newset = self.container._find_tracks(iter)
            tracks.update(newset)

        tracks = list(tracks)

        return trax.sort_tracks(common.BASE_SORT_TAGS, tracks)

    def on_query_tooltip(self, widget, x, y, keyboard_mode, tooltip):
        """
        Sets up a basic tooltip
        Required to have "&" in tooltips working
        """
        if not widget.get_tooltip_context(x, y, keyboard_mode):
            return False

        result = widget.get_path_at_pos(x, y)
        if not result:
            return False

        path = result[0]

        model = widget.get_model()
        tooltip.set_text(model[path][1])  # 1: title
        widget.set_tooltip_row(tooltip, path)

        return True

    def get_tracks_for_path(self, path):
        """
        Get tracks for a path from model (expand item)
        :param path: Gtk.TreePath
        :return: list of tracks [xl.trax.Track]
        """
        it = self.get_model().get_iter(path)
        search = self.container.get_node_search_terms(it)
        matcher = trax.TracksMatcher(search)
        for i in trax.search_tracks(self.container.tracks, [matcher]):
            yield i.track


# vim: et sts=4 sw=4



# Copyright (C) 2006 Adam Olsen
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 1, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

from gi.repository import GLib
from gi.repository import Gtk

import os.path
from . import _scrobbler

from xl import common, settings
from xl.nls import gettext as _
from xlgui import icons
from xlgui.preferences import widgets
from xlgui.widgets import dialogs

name = _('AudioScrobbler')
basedir = os.path.dirname(os.path.realpath(__file__))
ui = os.path.join(basedir, "asprefs_pane.ui")

icons.MANAGER.add_icon_name_from_directory(
    'audioscrobbler', os.path.join(basedir, 'icons')
)
icon = 'audioscrobbler'


class SubmitPreference(widgets.CheckPreference):
    default = True
    name = 'plugin/ascrobbler/submit'


class MenuCheck(widgets.CheckPreference):
    default = False
    name = 'plugin/ascrobbler/menu_check'


class ScrobbleRemote(widgets.CheckPreference):
    default = False
    name = 'plugin/ascrobbler/scrobble_remote'


class UserPreference(widgets.Preference):
    name = 'plugin/ascrobbler/user'


class PassPreference(widgets.HashedPreference):
    name = 'plugin/ascrobbler/password'


class UrlPreference(widgets.ComboEntryPreference):
    name = 'plugin/ascrobbler/url'
    default = 'http://post.audioscrobbler.com/'
    preset_items = {
        'http://post.audioscrobbler.com/': 'Last.fm',
        'http://turtle.libre.fm/': 'Libre.fm',
    }


class VerifyLoginButton(widgets.Button):
    name = 'plugin/ascrobbler/verify_login'

    def __init__(self, preferences, widget):
        """
        Sets up the message
        """
        widgets.Button.__init__(self, preferences, widget)

        self.message = dialogs.MessageBar(
            parent=preferences.builder.get_object('preferences_box'),
            buttons=Gtk.ButtonsType.CLOSE,
        )

    @common.threaded
    def check_login(self):
        """
        Tries to connect to the AudioScrobbler
        service with the existing login data
        """
        username = settings.get_option('plugin/ascrobbler/user', '')
        password = settings.get_option('plugin/ascrobbler/password', '')
        url = settings.get_option(
            'plugin/ascrobbler/url', 'http://post.audioscrobbler.com/'
        )
        login_verified = False

        try:
            _scrobbler.login(username, password, post_url=url)
        except _scrobbler.AuthError:
            try:
                _scrobbler.login(username, password, hashpw=True, post_url=url)
            except _scrobbler.AuthError:
                pass
            else:
                login_verified = True
        else:
            login_verified = True

        if login_verified:
            GLib.idle_add(self.message.show_info, _('Verification successful'), '')
        else:
            GLib.idle_add(
                self.message.show_error,
                _('Verification failed'),
                _('Please make sure the entered data is correct.'),
            )

        GLib.idle_add(self.widget.set_sensitive, True)

    def on_clicked(self, button):
        """
        Initiates verification of the login data
        """
        self.widget.set_sensitive(False)
        self.check_login()



# Copyright (C) 2012 Brian Parma
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 1, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

import os
import sys

from xlgui.preferences import widgets
from xl.nls import gettext as _

name = _('IPython Console')
basedir = os.path.dirname(os.path.realpath(__file__))
ui = os.path.join(basedir, 'ipconsole_prefs.ui')
icon = 'utilities-terminal'


class OpacityPreference(widgets.ScalePreference):
    default = 80.0
    name = 'plugin/ipconsole/opacity'

    def __init__(self, preferences, widget):
        widgets.ScalePreference.__init__(self, preferences, widget)
        if sys.platform.startswith("win32"):
            self.set_widget_sensitive(False)
            # Setting opacity on Windows crashes with segfault,
            # see https://bugzilla.gnome.org/show_bug.cgi?id=674449
            self.widget.set_tooltip(
                _("Opacity cannot be set on Windows due to a bug in Gtk+")
            )


class FontPreference(widgets.FontButtonPreference):
    default = 'Monospace 10'
    name = 'plugin/ipconsole/font'


class TextColor(widgets.RGBAButtonPreference):
    default = 'lavender'
    name = 'plugin/ipconsole/text_color'


class BgColor(widgets.RGBAButtonPreference):
    default = 'black'
    name = 'plugin/ipconsole/background_color'


class AutoStart(widgets.CheckPreference):
    default = False
    name = 'plugin/ipconsole/autostart'



# mpris2 - Support MPRIS 2 in Exaile
# Copyright (C) 2015-2016  Johannes Sasongko <sasongko@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
#
# The developers of the Exaile media player hereby grant permission
# for non-GPL compatible GStreamer and Exaile plugins to be used and
# distributed together with GStreamer and Exaile. This permission is
# above and beyond the permissions granted by the GPL license by which
# Exaile is covered. If you modify this code, you may extend this
# exception to your version of the code, but you are not obligated to
# do so. If you do not wish to do so, delete this exception statement
# from your version.


from gi.repository import GLib

Variant = GLib.Variant


class DBusHelper:
    """Helper to make D-Bus handling more Pythonic.

    The `object` argument of the constructor should be a Python object in the
    following form:

        class Object:
            def MyMethod(self, arg1, arg2):
                return GLib.Variant('s', 'out1'), GLib.Variant('s', 'out2')
            @property
            def MyProperty(self):
                return GLib.Variant('i', 0)
            @MyProperty.setter
            def MyProperty(self, value):
                pass

    You can then set the `method_call`, `get_property`, and `set_property`
    methods as the callbacks for the D-Bus object (i.e. the last three
    arguments in `DBusConnection.register_object`).

    Methods receive their arguments as plain Python objects, whereas property
    setters receive theirs as `GLib.Variant`s.
    Both methods and property setters must return their values as
    `GLib.Variant`s (technically, the return type is `Union[None, GLib.Variant,
    Tuple[GLib.Variant]]`).

    A Python method or property is exported if the name starts with an
    uppercase character and does not end with `_`.
    Attributes which are neither methods nor properties are not exported.
    """

    def __init__(self, obj):
        self.object = obj

    def method_call(
        self, connection, sender, path, interface, method, args, invocation
    ):
        self._check_method(method)
        result = getattr(self.object, method)(*args)
        # If the method returns nothing, return empty tuple. If the method
        # returns multiple values, return those. If the method returns a single
        # value, wrap in a tuple and return.
        if result is None:
            result = Variant.new_tuple()
        elif isinstance(result, tuple):
            assert all(isinstance(v, Variant) for v in result)
            result = Variant.new_tuple(*result)
        else:
            assert isinstance(result, Variant)
            result = Variant.new_tuple(result)
        invocation.return_value(result)

    def get_property(self, connection, sender, path, interface, prop):
        self._check_property(prop)
        return getattr(self.object, prop)

    def set_property(self, connection, sender, path, interface, prop, value):
        self._check_property(prop)
        setattr(self.object, prop, value)
        return True

    def _check_method(self, meth):
        """Check that `meth` is a valid property of `self.object`"""
        import types

        if meth and meth[0].isupper() and not meth.endswith('_'):
            classprop = getattr(self.object.__class__, meth, None)
            if isinstance(classprop, types.FunctionType):
                return
        raise AttributeError("Invalid method: " + meth)

    def _check_property(self, prop):
        """Check that `prop` is a valid property of `self.object`"""
        if prop and prop[0].isupper() and not prop.endswith('_'):
            classprop = getattr(self.object.__class__, prop, None)
            if isinstance(classprop, property):
                return
        raise AttributeError("Invalid property: " + prop)



from gi.repository import Gtk, GObject

import xlgui.main
import xl.main
from xl import event, settings, providers
from xl.nls import gettext as _

from . import qb_prefs


class QuickButtons:
    """
    Plugin adds some buttons on the bottom line of the playlists grid to
    change some settings quickly
    """

    name: str = 'QuickButtons'
    """
    Needed for MainWindowStatusBarPane
    """

    self_triggered: bool = False
    """
    Don't repeat yourself.
    Is set to True to prevent resetting from settings.set_option
    """

    _toolbar: Gtk.Box = None

    options = {
        "playlist/enqueue_by_default": {
            "show_button": "quickbuttons/btn_enqueue_by_default",
            "value": None,
            "default": False,
            "widget": None,
            "type": "toggle",
            "label": _("Enqueue"),
            "tooltip": _("Queue tracks by default instead of playing them"),
        },
        "queue/disable_new_track_when_playing": {
            "show_button": "quickbuttons/btn_disable_new_track_when_playing",
            "value": None,
            "default": False,
            "widget": None,
            "type": "toggle",
            "label": _("Keep playing"),
            "tooltip": _("Disallow playing new tracks when another track is playing"),
        },
        "queue/remove_item_when_played": {
            "show_button": "quickbuttons/btn_remove_item_when_played",
            "value": None,
            "default": True,
            "widget": None,
            "type": "toggle",
            "label": _("Auto-Remove"),
            "tooltip": _("Remove track from queue upon playback"),
        },
        "player/auto_advance": {
            "show_button": "quickbuttons/btn_auto_advance",
            "value": None,
            "default": True,
            "widget": None,
            "type": "toggle",
            "label": _("Auto-Advance"),
            "tooltip": _("Automatically advance to the next track"),
        },
        "player/auto_advance_delay": {
            "show_button": "quickbuttons/btn_auto_advance_delay",
            "value": None,
            "default": 0,
            "widget": None,
            "type": "spin",
            "label": _("Delay between tracks (in seconds):"),
            "tooltip": _("Delay between tracks (in seconds):"),
        },
        "equalizer": {
            "show_button": "quickbuttons/btn_equalizer",
            "value": None,
            "default": None,
            "widget": None,
            "type": "equalizer",
            "label": _("EQ"),
            "tooltip": _("Equalizer"),
            "depends_on": "equalizer",
        },
        "player/audiosink_device": {
            "show_button": "quickbuttons/btn_audio_device",
            "value": None,
            "default": 'auto',
            "widget": None,
            "type": "audio_device_selection",
            "label": _("Main Device"),
            "tooltip": _("Select main audio device"),
        },
        "preview_device/audiosink_device": {
            "show_button": "quickbuttons/btn_audio_device_preview",
            "value": None,
            "default": None,
            "widget": None,
            "type": "audio_device_selection",
            "label": _("Preview Device"),
            "tooltip": _("Select preview audio device"),
            "depends_on": "previewdevice",
        },
    }
    """
    Usable options
    """

    def enable(self, exaile: xl.main.Exaile):
        """
        Called on startup of exaile
        """
        self.exaile = exaile

    def disable(self, exaile: xl.main.Exaile):
        self._toolbar.hide()
        event.remove_callback(self._on_option_set, "playlist_option_set")
        event.remove_callback(self._on_option_set, "queue_option_set")
        event.remove_callback(self._on_option_set, "player_option_set")

    def _on_button_activate(self, event_name, event_source, option: str) -> None:
        for k in self.options:
            show_btn_option = self.options[k]["show_button"]
            if option == show_btn_option:
                if not self.options[k]["widget"]:
                    return
                if settings.get_option(show_btn_option, True):
                    self.options[k]["widget"].show_all()
                else:
                    self.options[k]["widget"].hide()
                break

    def _on_option_set(self, event_name, event_source, option: str) -> None:
        if option not in self.options:
            return

        if self.self_triggered:
            self.self_triggered = False
            return

        self.options[option]["value"] = settings.get_option(option)
        if self.options[option]["type"] == "toggle":
            self.options[option]["widget"].set_active(self.options[option]["value"])
        elif self.options[option]["type"] == "spin":
            self.options[option]["widget"].get_children()[0].set_value(
                self.options[option]["value"]
            )

    def _add_button(self, setting: str) -> None:
        if self.options[setting]["type"] == "toggle":
            tbs = qb_toggle(setting, self)

        elif self.options[setting]["type"] == "spin":
            tbs = qb_spinner(setting, self)

        elif self.options[setting]["type"] == "equalizer":
            tbs = qb_equalizer(setting, self)

        elif self.options[setting]["type"] == "audio_device_selection":
            tbs = qb_audio_device(setting, self)

        return self._add_button_to_toolbar(tbs, setting)

    def _add_button_to_toolbar(self, tbs: Gtk.Button, setting: str) -> None:
        show_btn_option = self.options[setting]["show_button"]
        show_btn = settings.get_option(show_btn_option, True)
        if show_btn:
            tbs.show_all()

        self.options[setting]["widget"] = tbs
        self._toolbar.pack_start(tbs, False, True, 0)

    def on_gui_loaded(self):
        """
        Called when the gui is loaded
        """
        if self._toolbar != None:
            self._toolbar.show()
            return

    def create_widget(self, info_area: xlgui.main.MainWindowStatusBarPane):
        self._toolbar = Gtk.Box()

        for k in self.options:
            self.options[k]["value"] = settings.get_option(
                k, self.options[k]["default"]
            )
            self._add_button(k)

        self._toolbar.show()
        return self._toolbar

    def on_exaile_loaded(self):
        providers.register('mainwindow-statusbar-widget', self)
        event.add_callback(self._on_option_set, "playlist_option_set")
        event.add_callback(self._on_option_set, "queue_option_set")
        event.add_callback(self._on_option_set, "player_option_set")
        event.add_callback(self._on_button_activate, "quickbuttons_option_set")

    def get_preferences_pane(self):
        return qb_prefs


plugin_class = QuickButtons


class qb_audio_device(Gtk.MenuButton):
    def __init__(self, setting, qb_instance):
        self._setting = setting
        self._qb = qb_instance
        self._settings = QuickButtons.options[setting]

        self._devices = {}
        """List of all known devices"""

        self._current_device = None
        """Currently selected device"""

        self._button_group = None
        """Button group of the radios"""

        self._devices_to = None
        """Timeout for _set_devices"""

        self.popover = Gtk.Popover()
        self.vbox = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
        self.vbox.show_all()
        self.popover.add(self.vbox)

        super().__init__(label=self._settings['label'], popover=self.popover)
        self.set_tooltip_text(self._settings["tooltip"])
        self.connect('toggled', self._on_cb_popup)

    def _set_audio_device(self, button: Gtk.RadioButton) -> None:
        self._qb.self_triggered = True
        if button.get_active() and button.device_id != self._current_device:
            settings.set_option(self._setting, button.device_id)

    def _on_cb_popup(self, widget: Gtk.RadioButton) -> None:
        if widget.get_active():
            self._set_devices()
        elif self._devices_to:
            GObject.source_remove(self._devices_to)

    def _set_devices(self) -> None:
        self._current_device = settings.get_option(
            self._setting, self._settings['default']
        )
        actual_devices = {}

        # @see plugins/previewdevice/previewprefs.py:65
        from xl.player.gst.sink import get_devices

        for name, device_id, _unused in reversed(list(get_devices())):
            actual_devices[device_id] = name
            if device_id not in self._devices:
                widget = self._add_toggle(name, device_id)
                self._devices[device_id] = {'name': name, 'widget': widget}

        for child in self.vbox.get_children():
            if child.device_id not in actual_devices:
                self._devices.pop(child.device_id, None)
                self.vbox.remove(child)

        if self._current_device not in actual_devices:
            self._current_device = 'auto'

        self._devices[self._current_device]["widget"].set_active(True)

        self.vbox.show_all()
        self.vbox.reorder_child(self._devices['auto']['widget'], -1)

        self._devices_to = GObject.timeout_add(1000, self._set_devices)

    def _add_toggle(self, label: str, device_id: str) -> None:
        btn = Gtk.RadioButton(label=label)
        btn.connect('toggled', self._set_audio_device)
        btn.device_id = device_id
        if self._button_group:
            btn.join_group(self._button_group)

        self._button_group = btn
        self.vbox.pack_start(btn, False, True, 10)
        return btn

    def show_all(self) -> None:
        if (
            "depends_on" in self._settings
            and self._settings['depends_on']
            not in self._qb.exaile.plugins.enabled_plugins
        ):
            super().hide()
        else:
            super().show_all()


class qb_equalizer(Gtk.Button):
    def __init__(self, setting, qb_instance):
        self._setting = setting
        self._qb = qb_instance
        self._settings = QuickButtons.options[setting]
        super().__init__()
        self.set_label(self._settings['label'])
        self.set_tooltip_text(self._settings["tooltip"])
        self.connect("clicked", self._on_equalizer_press)

    def _on_equalizer_press(self, widget) -> None:
        if "equalizer" not in self._qb.exaile.plugins.enabled_plugins:
            return None

        eq_plugin_win = self._qb.exaile.plugins.enabled_plugins["equalizer"].window
        if not eq_plugin_win:
            eq_win = GObject.new("EqualizerWindow")
            eq_win.set_transient_for(self._qb.exaile.gui.main.window)
            self._qb.exaile.plugins.enabled_plugins["equalizer"].window = eq_win
            self._equalizer_win = eq_win

        def _destroy(w):
            self._equalizer_win = None
            self._qb.exaile.plugins.enabled_plugins["equalizer"].window = None

        eq_win.connect("destroy", _destroy)
        eq_win.show_all()

    def show_all(self):
        if (
            "depends_on" in self._settings
            and self._settings['depends_on']
            not in self._qb.exaile.plugins.enabled_plugins
        ):
            super().hide()
        else:
            super().show_all()


class qb_spinner(Gtk.SpinButton):
    def __init__(self, setting, qb_instance: QuickButtons):
        self._setting = setting
        self._qb = qb_instance
        self._settings = QuickButtons.options[setting]
        super().__init__()
        self.set_tooltip_text(self._settings["tooltip"])
        self.set_adjustment(Gtk.Adjustment(self._get_delay_value(), 0, 60, 1, 0, 0))
        self.connect("value-changed", self._on_spin, setting)

    def _on_spin(self, widget, setting: str):
        """
        Called when changing the value from spinbutton
        """
        self._qb.self_triggered = True
        self._set_delay_value(widget.get_value_as_int())

    def _get_delay_value(self) -> int:
        """
        Get the current delay value in seconds as int
        """
        value = settings.get_option("player/auto_advance_delay")
        if value == None:
            value = 0
        value = value / 1000
        return int(value)

    def _set_delay_value(self, value: int) -> None:
        """
        Set the delay value in ms
        """
        value = value * 1000
        settings.set_option("player/auto_advance_delay", value)


class qb_toggle(Gtk.ToggleButton):
    def __init__(self, setting, qb_instance):
        self._setting = setting
        self._qb = qb_instance
        self._settings = QuickButtons.options[setting]
        active = self._settings["value"]
        if active != True:
            active = False
        super().__init__()
        self.set_label(self._settings["label"])
        self.set_tooltip_text(self._settings["tooltip"])
        self.set_active(active)
        self.connect("toggled", self._on_toggle, setting)

    def _on_toggle(self, widget, setting: str):
        """
        Called when toggling a button
        """
        self._qb.self_triggered = True
        settings.set_option(setting, widget.get_active())



# Copyright (C) 2009-2010 Dave Aitken
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
#
# The developers of the Exaile media player hereby grant permission
# for non-GPL compatible GStreamer and Exaile plugins to be used and
# distributed together with GStreamer and Exaile. This permission is
# above and beyond the permissions granted by the GPL license by which
# Exaile is covered. If you modify this code, you may extend this
# exception to your version of the code, but you are not obligated to
# do so. If you do not wish to do so, delete this exception statement
# from your version.


import logging
import math
import os

from gi.repository import Gst
from gi.repository import Gtk

from xl import providers, event, settings, xdg
from xl.player.gst.gst_utils import ElementBin
from xl.nls import gettext as _
from xlgui.guiutil import GtkTemplate
from xlgui.widgets import menu

logger = logging.getLogger('equalizer')


# Values from <http://www.xmms.org/faq.php#General3>, adjusted to be less loud
# in general ((mean + max) / 2 = 0).
DEFAULT_PRESETS = [
    # fmt: off
    ('Custom', 0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0),
    ('Default', 0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0),
    ('Classical', 0, 1.6, 1.6, 1.6, 1.6, 1.6, 1.6, -5.6, -5.6, -5.6, -8.0),
    ('Club', 0, -5.4, -5.4, 2.6, 0.2, 0.2, 0.2, -2.2, -5.4, -5.4, -5.4),
    ('Dance', 0, 4.8, 2.4, -2.4, -4.8, -4.8, -10.4, -12.0, -12.0, -4.8, -4.8),
    ('Full Bass', 0, -11.5, 6.1, 6.1, 2.1, -1.9, -7.5, -11.5, -13.9, -14.7, -14.7),
    ('Full Bass and Treble', 0, -1.1, -2.7, -8.3, -15.5, -13.1, -6.7, -0.3, 2.9, 3.7, 3.7),
    ('Full Treble', 0, -20.3, -20.3, -20.3, -14.7, -8.3, 0.5, 5.3, 5.3, 5.3, 6.1),
    ('Laptop Speakers and Headphones', 0, -3.9, 2.5, -3.1, -11.9, -11.1, -7.1, -3.9, 0.9, 3.2, 3.2),
    ('Large Hall', 0, 4.3, 4.3, -0.5, -0.5, -6.1, -10.9, -10.9, -10.9, -6.1, -6.1),
    ('Live', 0, -9.0, -4.2, -0.2, 1.4, 1.4, 1.4, -0.2, -1.8, -1.8, -1.8),
    ('Party', 0, 2.2, 2.2, -5.0, -5.0, -5.0, -5.0, -5.0, -5.0, 2.2, 2.2),
    ('Pop', 0, -6.4, 0.0, 2.4, 3.2, 0.8, -4.8, -7.2, -7.2, -6.4, -6.4),
    ('Reggae', 0, -3.6, -3.6, -3.6, -9.2, -3.6, 2.8, 2.8, -3.6, -3.6, -3.6),
    ('Rock', 0, 0.3, -2.9, -13.3, -15.7, -10.9, -3.7, 1.1, 3.5, 3.5, 3.5),
    ('Ska', 0, -9.9, -12.3, -11.5, -7.5, -3.5, -1.9, 1.3, 2.1, 3.7, 2.1),
    ('Soft', 0, -3.6, -6.8, -8.4, -10.8, -8.4, -4.4, -0.4, 1.2, 2.8, 3.6),
    ('Soft Rock', 0, -0.8, -0.8, -2.4, -4.8, -8.8, -10.4, -8.0, -4.8, -2.4, 4.0),
    ('Techno', 0, 1.2, -1.2, -6.8, -12.4, -11.6, -6.8, 1.2, 2.8, 2.8, 2.0),
    # fmt: on
]


class GSTEqualizer(ElementBin):
    """
    Equalizer GST class
    """

    index = 99
    name = "equalizer-10bands"

    def __init__(self):
        ElementBin.__init__(self, name=self.name)

        self.audioconvert = Gst.ElementFactory.make("audioconvert", None)
        self.elements[40] = self.audioconvert

        self.preamp = Gst.ElementFactory.make("volume", None)
        self.elements[50] = self.preamp

        self.eq10band = Gst.ElementFactory.make("equalizer-10bands", None)
        self.elements[60] = self.eq10band

        self.setup_elements()

        event.add_ui_callback(self._on_option_set, "plugin_equalizer_option_set")

        setts = ["band%s" for _number in range(10)] + ["pre", "enabled"]
        for setting in setts:
            self._on_option_set(
                "plugin_equalizer_option_set", None, "plugin/equalizer/%s" % setting
            )

    def _on_option_set(self, _name, _object, data):
        for band in range(10):
            if data == "plugin/equalizer/band%s" % band:
                if settings.get_option("plugin/equalizer/enabled") is True:
                    self.eq10band.set_property(
                        "band%s" % band,
                        settings.get_option("plugin/equalizer/band%s" % band),
                    )
                else:
                    self.eq10band.set_property("band%s" % band, 0.0)

        if data == "plugin/equalizer/pre":
            if settings.get_option("plugin/equalizer/enabled") is True:
                self.preamp.set_property(
                    "volume",
                    self.dB_to_percent(settings.get_option("plugin/equalizer/pre")),
                )
            else:
                self.preamp.set_property("volume", 1.0)

        if data == "plugin/equalizer/enabled":
            if settings.get_option("plugin/equalizer/enabled") is True:
                self.preamp.set_property(
                    "volume",
                    self.dB_to_percent(settings.get_option("plugin/equalizer/pre")),
                )
                for band in range(10):
                    self.eq10band.set_property(
                        "band%s" % band,
                        settings.get_option("plugin/equalizer/band%s" % band),
                    )
            else:
                self.preamp.set_property("volume", 1.0)
                for band in range(10):
                    self.eq10band.set_property("band%s" % band, 0.0)

    @staticmethod
    def dB_to_percent(dB):
        return 10 ** (dB / 10)


@GtkTemplate('equalizer.ui', relto=__file__)
class EqualizerWindow(Gtk.Window):
    __gtype_name__ = 'EqualizerWindow'

    PRESETS_PATH = os.path.join(xdg.get_config_dir(), 'eq-presets.dat')

    (
        band0,
        band1,
        band2,
        band3,
        band4,
        band5,
        band6,
        band7,
        band8,
        band9,
        chk_enabled,
        combo_presets,
        presets,
        pre,
    ) = GtkTemplate.Child.widgets(14)

    def __init__(self):
        Gtk.Window.__init__(self)
        self.init_template()
        self.pre.set_value(settings.get_option("plugin/equalizer/pre"))
        self.chk_enabled.set_active(settings.get_option("plugin/equalizer/enabled"))
        # Setup bands/preamp from current equalizer settings
        for number in (0, 1, 2, 3, 4, 5, 6, 7, 8, 9):
            band = getattr(self, 'band%s' % number)
            band.set_value(self.get_band(number))
        self.combo_presets.set_entry_text_column(0)
        self.combo_presets.set_active(0)
        self.load_presets()

    @staticmethod
    def get_band(band_number):
        """
        Get the current value of band x
        """
        return settings.get_option("plugin/equalizer/band%s" % band_number)

    @GtkTemplate.Callback
    def adjust_band(self, widget):
        """
        Adjust the specified band
        """
        # Buildable.get_name clashes with Widget.get_name. See
        # https://bugzilla.gnome.org/show_bug.cgi?id=591085#c19
        widget_name = Gtk.Buildable.get_name(widget)
        band = widget_name[-1]
        settings_value = settings.get_option("plugin/equalizer/band" + band)
        if not math.isclose(widget.get_value(), settings_value):
            settings.set_option("plugin/equalizer/band" + band, widget.get_value())
            self.combo_presets.set_active(0)

    @GtkTemplate.Callback
    def adjust_preamp(self, widget):
        """
        Adjust the preamp
        """
        if widget.get_value() != settings.get_option("plugin/equalizer/pre"):
            settings.set_option("plugin/equalizer/pre", widget.get_value())
            self.combo_presets.set_active(0)

    @GtkTemplate.Callback
    def add_preset(self, _widget):
        new_preset = []
        new_preset.append(self.combo_presets.get_child().get_text())
        new_preset.append(settings.get_option("plugin/equalizer/pre"))

        for band in range(10):
            new_preset.append(settings.get_option("plugin/equalizer/band%s" % band))

        self.presets.append(new_preset)
        self.save_presets()

    @GtkTemplate.Callback
    def remove_preset(self, _widget):
        entry = self.combo_presets.get_active()
        if entry > 1:
            self.presets.remove(self.presets.get_iter(entry))
            self.combo_presets.set_active(0)
            self.save_presets()

    @GtkTemplate.Callback
    def preset_changed(self, widget):
        model = widget.get_model()
        index = widget.get_active()

        # If an option other than "Custom" is chosen:
        if index > 0:
            settings.set_option(
                "plugin/equalizer/pre", model.get_value(model.get_iter(index), 1)
            )
            self.pre.set_value(model.get_value(model.get_iter(index), 1))

            for band in range(10):
                settings.set_option(
                    "plugin/equalizer/band%s" % band,
                    model.get_value(model.get_iter(index), band + 2),
                )
                band_widget = getattr(self, "band%s" % band)
                band_widget.set_value(model.get_value(model.get_iter(index), band + 2))

    @GtkTemplate.Callback
    def check_enabled(self, widget):
        settings.set_option("plugin/equalizer/enabled", widget.get_active())

    def save_presets(self):
        if os.path.exists(self.PRESETS_PATH):
            os.remove(self.PRESETS_PATH)

        with open(self.PRESETS_PATH, 'w') as config_file:
            for row in self.presets:
                config_file.write(row[0] + '\n')
                line = "".join(str(row[i]) + " " for i in range(1, 12))
                line += "\n"
                config_file.write(line)

    def load_presets(self):
        """
        Populate the GTK ListStore with presets
        """
        load_defaults = True
        if os.path.exists(self.PRESETS_PATH):
            try:
                with open(self.PRESETS_PATH, 'r') as presets_file:
                    line = presets_file.readline()
                    while line != "":
                        preset = []
                        preset.append(line[:-1])
                        line = presets_file.readline()
                        vals = line.split(" ")
                        for i in range(11):
                            preset.append(float(vals[i]))

                        self.presets.append(preset)
                        line = presets_file.readline()
            except Exception:
                logger.exception(
                    "Error loading equalizer presets, reverting to defaults"
                )
            else:
                load_defaults = False

        if load_defaults:
            for preset in DEFAULT_PRESETS:
                self.presets.append(preset)


class EqualizerPlugin:
    """
    Equalizer plugin class
    """

    def __init__(self):
        self.window = None
        self.__menu_item = None
        self.__exaile = None
        self.check_default_settings()

    @staticmethod
    def check_default_settings():
        for band in range(10):
            if settings.get_option("plugin/equalizer/band%s" % band) is None:
                settings.set_option("plugin/equalizer/band%s" % band, 0.0)

        if settings.get_option("plugin/equalizer/pre") is None:
            settings.set_option("plugin/equalizer/pre", 0.0)

        if settings.get_option("plugin/equalizer/enabled") is None:
            settings.set_option("plugin/equalizer/enabled", True)

    def disable(self, _exaile):
        if self.__menu_item:
            providers.unregister('menubar-tools-menu', self.__menu_item)
            self.__menu_item = None
        if self.window:
            self.window.hide()
            self.window.destroy()
            self.window = None
        providers.unregister("gst_audio_filter", GSTEqualizer)

    def __show_gui(self):
        """
        Display main window.
        """
        if not self.window:
            self.window = EqualizerWindow()
            self.window.set_transient_for(self.__exaile.gui.main.window)

            def _destroy(w):
                self.window = None

            self.window.connect('destroy', _destroy)
        self.window.show_all()

    def enable(self, exaile):
        providers.register("gst_audio_filter", GSTEqualizer)
        self.__exaile = exaile

    def on_gui_loaded(self):
        """
        Called when the player is loaded.
        """
        # add menu item to tools menu
        self.__menu_item = menu.simple_menu_item(
            'equalizer',
            ['plugin-sep'],
            _('_Equalizer'),
            callback=lambda *x: self.__show_gui(),
        )
        providers.register('menubar-tools-menu', self.__menu_item)


plugin_class = EqualizerPlugin



# Copyright (C) 2024  Johannes Sasongko <johannes sasongko org>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
#
# The developers of the Exaile media player hereby grant permission
# for non-GPL compatible GStreamer and Exaile plugins to be used and
# distributed together with GStreamer and Exaile. This permission is
# above and beyond the permissions granted by the GPL license by which
# Exaile is covered. If you modify this code, you may extend this
# exception to your version of the code, but you are not obligated to
# do so. If you do not wish to do so, delete this exception statement
# from your version.


import io
import os
import struct
from typing import Tuple

from xl.metadata._base import BaseFormat, NotReadable


class AuFormat(BaseFormat):
    writable = False

    def load(self):
        try:
            with open(self.loc, 'rb', buffering=0) as f:
                file_size = os.stat(f.fileno()).st_size
                bitrate, length = _read_metadata(f, file_size)
        except IOError:
            raise NotReadable
        self.mutagen = {'__bitrate': bitrate, '__length': length}


def _read_metadata(f: io.FileIO, file_size: int) -> Tuple[float, float]:
    """
    :return: bitrate, length
    """

    # See https://en.wikipedia.org/wiki/Au_file_format

    try:
        header = f.read(24)
    except IOError:
        return -1, -1

    # > = big-endian; I = uint32
    magic, data_offset, data_size, encoding, sample_rate, channels = struct.unpack(
        '>IIIIII', header
    )
    if magic != 0x2E736E64:  # '.snd'
        return -1, -1

    # Data size can be unknown (uint32_max) or wrong; try to guess it
    data_size = min(data_size, file_size - data_offset)

    # fmt: off
    ENCODING_WIDTHS = {
        1: 1,
        2: 1, 3: 2, 4: 3, 5: 4,
        6: 4, 7: 8,
        11: 1, 12: 2, 13: 3, 14: 4,
        18: 2, 19: 2, 20: 2,
        27: 1,
    }
    # fmt: on
    try:
        sample_width = ENCODING_WIDTHS[encoding]
    except KeyError:
        return -1, -1

    byterate = sample_rate * sample_width * channels
    return byterate * 8, data_size / byterate



# Copyright (C) 2008-2010 Adam Olsen
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
#
# The developers of the Exaile media player hereby grant permission
# for non-GPL compatible GStreamer and Exaile plugins to be used and
# distributed together with GStreamer and Exaile. This permission is
# above and beyond the permissions granted by the GPL license by which
# Exaile is covered. If you modify this code, you may extend this
# exception to your version of the code, but you are not obligated to
# do so. If you do not wish to do so, delete this exception statement
# from your version.


import logging
from typing import Dict, Union

from gi.repository import Gtk
from gi.repository import Gdk
from gi.repository import GdkPixbuf
from gi.repository import GObject

from xl import common, event, radio, settings, trax
from xl.nls import gettext as _
from xl.playlist import Playlist, SmartPlaylist
from xlgui import icons, panel
from xlgui.panel import menus
from xlgui.widgets import dialogs
from xlgui.widgets.common import DragTreeView
from xlgui.widgets.smart_playlist_editor import SmartPlaylistEditor

logger = logging.getLogger(__name__)


class TrackWrapper:
    def __init__(self, track, playlist):
        self.track = track
        self.playlist = playlist

    def __str__(self):
        text = self.track.get_tag_raw('title')
        if text is not None:
            text = ' / '.join(text)

        if text:
            artists = self.track.get_tag_raw('artist')
            if artists:
                text += ' - ' + ' / '.join(artists)
            return text
        return self.track.get_loc_for_io()


class BasePlaylistPanelMixin(GObject.Object):
    """
    Base playlist tree object.

    Used by the radio and playlists panels to display playlists
    """

    # HACK: Notice that this is not __gsignals__; descendants need to manually
    # merge this in. This is because new PyGObject doesn't like __gsignals__
    # coming from mixin. See:
    # * https://bugs.launchpad.net/bugs/714484
    # * http://www.daa.com.au/pipermail/pygtk/2011-February/019394.html
    _gsignals_ = {
        'playlist-selected': (GObject.SignalFlags.RUN_LAST, None, (object,)),
        'tracks-selected': (GObject.SignalFlags.RUN_LAST, None, (object,)),
        'append-items': (GObject.SignalFlags.RUN_LAST, None, (object, bool)),
        'replace-items': (GObject.SignalFlags.RUN_LAST, None, (object,)),
        'queue-items': (GObject.SignalFlags.RUN_LAST, None, (object,)),
    }

    # Cache for custom playlists
    playlist_nodes: Dict[Playlist, Gtk.TreeIter]

    # Mapping to keep track of open "are you sure you want to delete" dialogs
    deletion_dialogs: Dict[Union[Playlist, SmartPlaylist], Gtk.Dialog]

    def __init__(self):
        """
        Initializes the mixin
        """
        GObject.Object.__init__(self)
        self.playlist_nodes = {}
        self.track_image = icons.MANAGER.pixbuf_from_icon_name(
            'audio-x-generic', Gtk.IconSize.SMALL_TOOLBAR
        )
        self.deletion_dialogs = {}

    def remove_playlist(self, ignored=None):
        """
        Removes the selected playlist from the UI
        and from the underlying manager
        """
        selected_playlist = self.tree.get_selected_page(raw=True)
        if selected_playlist is None:
            return
        dialog = self.deletion_dialogs.get(selected_playlist)
        if dialog:
            dialog.present()
            return

        def on_response(dialog, response):
            if response == Gtk.ResponseType.YES:
                if isinstance(selected_playlist, SmartPlaylist):
                    self.smart_manager.remove_playlist(selected_playlist.name)
                else:
                    self.playlist_manager.remove_playlist(selected_playlist.name)
                    # Remove from {playlist: iter} cache.
                    del self.playlist_nodes[selected_playlist]
                # Remove from UI.
                selection = self.tree.get_selection()
                model, iter = selection.get_selected()
                self.model.remove(iter)
            del self.deletion_dialogs[selected_playlist]
            dialog.destroy()

        dialog = Gtk.MessageDialog(
            buttons=Gtk.ButtonsType.YES_NO,
            message_type=Gtk.MessageType.QUESTION,
            text=_('Delete the playlist "%s"?') % selected_playlist.name,
            transient_for=self.parent,
        )
        dialog.connect('response', on_response)
        self.deletion_dialogs[selected_playlist] = dialog
        dialog.present()

    def rename_playlist(self, playlist):
        """
        Renames the playlist
        """

        if playlist is None:
            return

        # Ask for new name
        dialog = dialogs.TextEntryDialog(
            _("Enter the new name you want for your playlist"),
            _("Rename Playlist"),
            playlist.name,
            parent=self.parent,
        )

        result = dialog.run()
        name = dialog.get_value()

        dialog.destroy()

        if result != Gtk.ResponseType.OK or name == '':
            return

        if name in self.playlist_manager.playlists:
            # name is already in use
            dialogs.error(
                self.parent, _("The playlist name you entered is already in use.")
            )
            return

        selection = self.tree.get_selection()
        model, iter = selection.get_selected()
        model.set_value(iter, 1, name)

        # Update the manager as well
        self.playlist_manager.rename_playlist(playlist, name)

    def open_selected_playlist(self):
        selection = self.tree.get_selection()
        model, iter = selection.get_selected()
        self.open_item(self.tree, model.get_path(iter), None)

    def on_rating_changed(self, widget, rating):
        """
        Updates the rating of the selected tracks
        """
        tracks = self.get_selected_tracks()

        for track in tracks:
            track.set_rating(rating)

        maximum = settings.get_option('rating/maximum', 5)
        event.log_event('rating_changed', self, 100 * rating / maximum)

    def open_item(self, tree, path, col):
        """
        Called when the user double clicks on a playlist,
        also called when the user double clicks on a track beneath
        a playlist.  When they active a track it opens the playlist
        and starts playing that track
        """
        iter = self.model.get_iter(path)
        item = self.model.get_value(iter, 2)
        if item is not None:
            if isinstance(item, (Playlist, SmartPlaylist)):
                # for smart playlists
                if hasattr(item, 'get_playlist'):
                    try:
                        item = item.get_playlist(self.collection)
                    except Exception as e:
                        logger.exception("Error loading smart playlist")
                        dialogs.error(
                            self.parent, _("Error loading smart playlist: %s") % str(e)
                        )
                        return
                else:
                    # Get an up to date copy
                    item = self.playlist_manager.get_playlist(item.name)
                    # item.set_is_custom(True)

                #                self.controller.main.add_playlist(item)
                self.emit('playlist-selected', item)
            else:
                self.emit('append-items', [item.track], True)

    def add_new_playlist(self, tracks=[], name=None):
        """
        Adds a new playlist to the list of playlists. If name is
        None or the name conflicts with an existing playlist, the
        user will be queried for a new name.

        Returns the name of the new playlist, or None if it was
        not added.
        """
        if name:
            if name in self.playlist_manager.playlists:
                name = dialogs.ask_for_playlist_name(
                    self.get_panel().get_toplevel(), self.playlist_manager, name
                )
        else:
            if tracks:
                artists = []
                composers = []
                albums = []

                for track in tracks:
                    artist = track.get_tag_display('artist', artist_compilations=False)

                    if artist is not None:
                        artists += [artist]

                    composer = track.get_tag_display(
                        'composer', artist_compilations=False
                    )

                    if composer is not None:
                        composers += composer

                    album = track.get_tag_display('album')

                    if album is not None:
                        albums += album

                artists = list(set(artists))[:3]
                composers = list(set(composers))[:3]
                albums = list(set(albums))[:3]

                if len(artists) > 0:
                    name = artists[0]

                    if len(artists) > 2:
                        # TRANSLATORS: Playlist title suggestion with more
                        # than two values
                        name = _('%(first)s, %(second)s and others') % {
                            'first': artists[0],
                            'second': artists[1],
                        }
                    elif len(artists) > 1:
                        # TRANSLATORS: Playlist title suggestion with two values
                        name = _('%(first)s and %(second)s') % {
                            'first': artists[0],
                            'second': artists[1],
                        }
                elif len(composers) > 0:
                    name = composers[0]

                    if len(composers) > 2:
                        # TRANSLATORS: Playlist title suggestion with more
                        # than two values
                        name = _('%(first)s, %(second)s and others') % {
                            'first': composers[0],
                            'second': composers[1],
                        }
                    elif len(composers) > 1:
                        # TRANSLATORS: Playlist title suggestion with two values
                        name = _('%(first)s and %(second)s') % {
                            'first': composers[0],
                            'second': composers[1],
                        }
                elif len(albums) > 0:
                    name = albums[0]

                    if len(albums) > 2:
                        # TRANSLATORS: Playlist title suggestion with more
                        # than two values
                        name = _('%(first)s, %(second)s and others') % {
                            'first': albums[0],
                            'second': albums[1],
                        }
                    elif len(albums) > 1:
                        # TRANSLATORS: Playlist title suggestion with two values
                        name = _('%(first)s and %(second)s') % {
                            'first': albums[0],
                            'second': albums[1],
                        }
                else:
                    name = ''

            name = dialogs.ask_for_playlist_name(
                self.get_panel().get_toplevel(), self.playlist_manager, name
            )

        if name is not None:
            # Create the playlist from all of the tracks
            new_playlist = Playlist(name)
            new_playlist.extend(tracks)
            # We are adding a completely new playlist with tracks so we save it
            self.playlist_manager.save_playlist(new_playlist)

        return name

    def _load_playlist_nodes(self, playlist):
        """
        Loads the playlist tracks into the node for the specified playlist
        """
        if playlist not in self.playlist_nodes:
            return

        expanded = self.tree.row_expanded(
            self.model.get_path(self.playlist_nodes[playlist])
        )

        self._clear_node(self.playlist_nodes[playlist])
        parent = self.playlist_nodes[playlist]
        for track in playlist:
            if not track:
                continue
            wrapper = TrackWrapper(track, playlist)
            row = (self.track_image, str(wrapper), wrapper)
            self.model.append(parent, row)

        if expanded:
            self.tree.expand_row(
                self.model.get_path(self.playlist_nodes[playlist]), False
            )

    def remove_selected_track(self):
        """
        Removes the selected track from its playlist
        and saves the playlist
        """
        selection = self.tree.get_selection()
        model, iter = selection.get_selected()
        track = model.get_value(iter, 2)
        if isinstance(track, TrackWrapper):
            del track.playlist[track.playlist.index(track.track)]
            # Update the list
            self.model.remove(iter)
            # TODO do we save the playlist after this??
            self.playlist_manager.save_playlist(track.playlist, overwrite=True)


class PlaylistsPanel(panel.Panel, BasePlaylistPanelMixin):
    """
    The playlists panel
    """

    __gsignals__ = BasePlaylistPanelMixin._gsignals_

    ui_info = ('playlists.ui', 'PlaylistsPanel')

    def __init__(self, parent, playlist_manager, smart_manager, collection, name):
        """
        Initializes the playlists panel

        @param playlist_manager:  The playlist manager
        """
        panel.Panel.__init__(self, parent, name, _('Playlists'))
        BasePlaylistPanelMixin.__init__(self)
        self.playlist_manager = playlist_manager
        self.smart_manager = smart_manager
        self.collection = collection
        self.box = self.builder.get_object('PlaylistsPanel')

        self.playlist_name_info = 500
        self.track_target = Gtk.TargetEntry.new("text/uri-list", 0, 0)
        self.playlist_target = Gtk.TargetEntry.new(
            "playlist_name", Gtk.TargetFlags.SAME_WIDGET, self.playlist_name_info
        )
        self.deny_targets = [Gtk.TargetEntry.new('', 0, 0)]

        self.tree = PlaylistDragTreeView(self)
        self.tree.connect('row-activated', self.open_item)
        self.tree.set_headers_visible(False)
        self.tree.connect('drag-motion', self.drag_motion)
        self.tree.drag_source_set(
            Gdk.ModifierType.BUTTON1_MASK,
            [self.track_target, self.playlist_target],
            Gdk.DragAction.COPY | Gdk.DragAction.MOVE,
        )

        self.scroll = Gtk.ScrolledWindow()
        self.scroll.set_policy(Gtk.PolicyType.AUTOMATIC, Gtk.PolicyType.AUTOMATIC)
        self.scroll.add(self.tree)
        self.scroll.set_shadow_type(Gtk.ShadowType.IN)
        self.box.pack_start(self.scroll, True, True, 0)
        self.box.show_all()

        pb = Gtk.CellRendererPixbuf()
        cell = Gtk.CellRendererText()
        if settings.get_option('gui/ellipsize_text_in_panels', False):
            from gi.repository import Pango

            cell.set_property('ellipsize-set', True)
            cell.set_property('ellipsize', Pango.EllipsizeMode.END)
        col = Gtk.TreeViewColumn('Text')
        col.pack_start(pb, False)
        col.pack_start(cell, True)
        col.set_attributes(pb, pixbuf=0)
        col.set_attributes(cell, text=1)
        self.tree.append_column(col)
        self.model = Gtk.TreeStore(GdkPixbuf.Pixbuf, str, object)
        self.tree.set_model(self.model)

        # icons
        self.folder = icons.MANAGER.pixbuf_from_icon_name(
            'folder', Gtk.IconSize.SMALL_TOOLBAR
        )
        self.playlist_image = icons.MANAGER.pixbuf_from_icon_name(
            'music-library', Gtk.IconSize.SMALL_TOOLBAR
        )

        # menus
        self.playlist_menu = menus.PlaylistsPanelPlaylistMenu(self)
        self.smart_menu = menus.PlaylistsPanelPlaylistMenu(self)
        self.default_menu = menus.PlaylistPanelMenu(self)

        self.track_menu = menus.TrackPanelMenu(self)

        self._connect_events()
        self._load_playlists()

    @property
    def menu(self):
        """
        Gets a menu for the selected item
        :return: xlgui.widgets.menu.Menu or None if do not have it
        """
        model, it = self.tree.get_selection().get_selected()
        pl = model[it][2]
        return (
            self.playlist_menu
            if isinstance(pl, Playlist)
            else (
                self.smart_menu
                if isinstance(pl, SmartPlaylist)
                else (
                    self.track_menu
                    if isinstance(pl, TrackWrapper)
                    else self.default_menu
                )
            )
        )

    def _connect_events(self):
        event.add_ui_callback(self.refresh_playlists, 'track_tags_changed')
        event.add_ui_callback(
            self._on_playlist_added, 'playlist_added', self.playlist_manager
        )

        self.tree.connect('key-release-event', self.on_key_released)

    def _playlist_properties(self):
        pl = self.tree.get_selected_page(raw=True)
        if isinstance(pl, SmartPlaylist):
            self.edit_selected_smart_playlist()

    def refresh_playlists(self, type, track, tags):
        """
        wrapper so that multiple events dont cause multiple
        reloads in quick succession
        """
        if settings.get_option('gui/sync_on_tag_change', True) and tags & {
            'title',
            'artist',
        }:
            self._refresh_playlists()

    @common.glib_wait(500)
    def _refresh_playlists(self):
        """
        Callback for when tags have changed and the playlists
        need refreshing.
        """
        if settings.get_option('gui/sync_on_tag_change', True):
            for playlist in self.playlist_nodes:
                self._load_playlist_nodes(playlist)

    def _on_playlist_added(self, type, object, playlist_name):
        new_playlist = self.playlist_manager.get_playlist(playlist_name)

        for oldpl in self.playlist_nodes:
            if oldpl.name == playlist_name:  # Name already exists
                if oldpl is not new_playlist:
                    node = self.playlist_nodes[oldpl]
                    # Replace the playlist object in {playlist: iter} cache.
                    del self.playlist_nodes[oldpl]
                    self.playlist_nodes[new_playlist] = node
                    # Replace the playlist object in tree model.
                    self.model[node][2] = new_playlist
                break
        else:  # Name doesn't exist yet
            self.playlist_nodes[new_playlist] = self.model.append(
                self.custom, [self.playlist_image, playlist_name, new_playlist]
            )
            self.tree.expand_row(self.model.get_path(self.custom), False)

        # Refresh the playlist subnodes.
        self._load_playlist_nodes(new_playlist)

    def _load_playlists(self):
        """
        Loads the currently saved playlists
        """
        self.smart = self.model.append(None, [self.folder, _("Smart Playlists"), None])

        self.custom = self.model.append(
            None, [self.folder, _("Custom Playlists"), None]
        )

        names = sorted(self.smart_manager.playlists)
        for name in names:
            self.model.append(
                self.smart,
                [self.playlist_image, name, self.smart_manager.get_playlist(name)],
            )

        names = sorted(self.playlist_manager.playlists)
        for name in names:
            playlist = self.playlist_manager.get_playlist(name)
            self.playlist_nodes[playlist] = self.model.append(
                self.custom, [self.playlist_image, name, playlist]
            )
            self._load_playlist_nodes(playlist)

        self.tree.expand_row(self.model.get_path(self.smart), False)
        self.tree.expand_row(self.model.get_path(self.custom), False)

    def import_playlist(self):
        """
        Shows a dialog to ask the user to import a new playlist
        """

        def _on_playlists_selected(dialog, playlists):
            for playlist in playlists:
                self.add_new_playlist(playlist, playlist.name)

        dialog = dialogs.PlaylistImportDialog(parent=self.parent)
        dialog.connect('playlists-selected', _on_playlists_selected)
        dialog.show()

    def add_smart_playlist(self):
        """
        Shows a dialog for adding a new smart playlist
        """
        pl = SmartPlaylistEditor.create(
            self.collection, self.smart_manager, self.parent
        )
        if pl:
            self.model.append(self.smart, [self.playlist_image, pl.name, pl])

    def edit_selected_smart_playlist(self):
        """
        Shows a dialog for editing the currently selected smart playlist
        """
        pl = self.tree.get_selected_page(raw=True)
        self.edit_smart_playlist(pl)

    def edit_smart_playlist(self, pl):
        """
        Shows a dialog for editing a smart playlist
        """
        pl = SmartPlaylistEditor.edit(
            pl, self.collection, self.smart_manager, self.parent
        )
        if pl:
            selection = self.tree.get_selection()
            model, it = selection.get_selected()
            model.set_value(it, 1, pl.name)
            model.set_value(it, 2, pl)

    def drag_data_received(self, tv, context, x, y, selection, info, etime):
        """
        Called when someone drags some thing onto the playlist panel
        """
        if info == self.playlist_name_info:
            # We are being dragged a playlist so
            # we have to reorder them
            playlist_name = selection.get_text()
            drag_source = self.tree.get_selected_page()
            # verify names
            if drag_source is not None:
                if drag_source.name == playlist_name:
                    drop_info = tv.get_dest_row_at_pos(x, y)
                    drag_source_iter = self.playlist_nodes[drag_source]
                    if drop_info:
                        path, position = drop_info
                        drop_target_iter = self.model.get_iter(path)
                        drop_target = self.model.get_value(drop_target_iter, 2)
                        if position == Gtk.TreeViewDropPosition.BEFORE:
                            # Put the playlist before drop_target
                            self.model.move_before(drag_source_iter, drop_target_iter)
                            self.playlist_manager.move(
                                playlist_name, drop_target.name, after=False
                            )
                        else:
                            # put the playlist after drop_target
                            self.model.move_after(drag_source_iter, drop_target_iter)
                            self.playlist_manager.move(
                                playlist_name, drop_target.name, after=True
                            )
            # Even though we are doing a move we still don't
            # call the delete method because we take care
            # of it above by moving instead of inserting/deleting
            context.finish(True, False, etime)
        else:
            self._drag_data_received_uris(tv, context, x, y, selection, info, etime)

    def _drag_data_received_uris(self, tv, context, x, y, selection, info, etime):
        """
        Called by drag_data_received when the user drags URIs onto us
        """
        locs = list(selection.get_uris())
        drop_info = tv.get_dest_row_at_pos(x, y)
        if drop_info:
            path, position = drop_info
            iter = self.model.get_iter(path)
            drop_target = self.model.get_value(iter, 2)

            # if the current item is a track, use the parent playlist
            insert_index = None
            if isinstance(drop_target, TrackWrapper):
                current_playlist = drop_target.playlist
                drop_target_index = current_playlist.index(drop_target.track)
                # Adjust insert position based on drop position
                if (
                    position == Gtk.TreeViewDropPosition.BEFORE
                    or position == Gtk.TreeViewDropPosition.INTO_OR_BEFORE
                ):
                    # By default adding tracks inserts it before so we do not
                    # have to modify the insert index
                    insert_index = drop_target_index
                else:
                    # If we want to go after we have to append 1
                    insert_index = drop_target_index + 1
            else:
                current_playlist = drop_target

            # Since the playlist do not have very good support for
            # duplicate tracks we have to perform some trickery
            # to make this work properly in all cases
            try:
                remove_track_index = current_playlist.index(
                    self.tree.get_selected_track()
                )
            except ValueError:
                remove_track_index = None
            if insert_index is not None and remove_track_index is not None:
                # Since remove_track_index will be removed before
                # the new track is inserted we have to offset the
                # insert index
                if insert_index > remove_track_index:
                    insert_index = insert_index - 1

            # Delete the track before adding the other one
            # so we do not get duplicates
            # right now the playlist does not support
            # duplicate tracks very well
            if context.get_selected_action() == Gdk.DragAction.MOVE:
                # On a move action the second True makes the
                # drag_data_delete function called
                context.finish(True, True, etime)
            else:
                context.finish(True, False, etime)

            # Add the tracks we found to the internal playlist
            # TODO: have it pass in existing tracks?
            tracks, playlists = self.tree.get_drag_data(locs)

            if insert_index is not None:
                current_playlist[insert_index:insert_index] = tracks
            else:
                current_playlist.extend(tracks)

            self._load_playlist_nodes(current_playlist)

            # Do we save in the case when a user drags a file onto a playlist
            # in the playlist panel? note that the playlist does not have to
            # be open for this to happen
            self.playlist_manager.save_playlist(current_playlist, overwrite=True)
        else:
            # If the user dragged files prompt for a new playlist name
            # else if they dragged a playlist add the playlist

            # We don't want the tracks in the playlists to be added to the
            # master tracks list so we pass in False
            tracks, playlists = self.tree.get_drag_data(locs, False)
            # First see if they dragged any playlist files
            for new_playlist in playlists:
                # We are adding a completely new playlist with tracks so
                # we save it. This will trigger playlist_added.
                self.playlist_manager.save_playlist(new_playlist, overwrite=True)

            # After processing playlist proceed to ask the user for the
            # name of the new playlist to add and add the tracks to it
            if len(tracks) > 0:
                self.add_new_playlist(tracks)

    def drag_data_delete(self, tv, context):
        """
        Called after a drag data operation is complete
        and we want to delete the source data
        """
        if Gdk.drag_drop_succeeded(context):
            self.remove_selected_track()

    def drag_get_data(self, tv, context, selection_data, info, time):
        """
        Called when someone drags something from the playlist
        """
        # TODO based on info determine what we set in selection_data
        if info == self.playlist_name_info:
            pl = self.tree.get_selected_page()
            if pl is not None:
                selection_data.set_text(pl.name, len(pl.name))
        else:
            pl = self.tree.get_selected_page()
            if pl is not None:
                tracks = pl[:]
            else:
                tracks = self.tree.get_selected_tracks()

            if not tracks:
                return

            for track in tracks:
                DragTreeView.dragged_data[track.get_loc_for_io()] = track

            uris = trax.util.get_uris_from_tracks(tracks)
            selection_data.set_uris(uris)

    def drag_motion(self, tv, context, x, y, time):
        """
        Sets the appropriate drag action based on what we are hovering over

        hovering over playlists causes the copy action to occur
        hovering over tracks within the same playlist causes the move
            action to occur
        hovering over tracks within different playlist causes the move
            action to occur

        Called on the destination widget
        """
        # Reset any target to be default to moving tracks
        self.tree.enable_model_drag_dest([self.track_target], Gdk.DragAction.DEFAULT)
        # Determine where the drag is coming from
        dragging_playlist = False
        if tv == self.tree:
            selected_playlist = self.tree.get_selected_page()
            if selected_playlist is not None:
                dragging_playlist = True

        # Find out where they are dropping onto
        drop_info = tv.get_dest_row_at_pos(x, y)
        if drop_info:
            path, position = drop_info
            iter = self.model.get_iter(path)
            drop_target = self.model.get_value(iter, 2)

            if isinstance(drop_target, Playlist):
                if dragging_playlist:
                    # If we drag onto  we copy, if we drag between we move
                    if (
                        position == Gtk.TreeViewDropPosition.INTO_OR_BEFORE
                        or position == Gtk.TreeViewDropPosition.INTO_OR_AFTER
                    ):
                        Gdk.drag_status(context, Gdk.DragAction.COPY, time)
                    else:
                        Gdk.drag_status(context, Gdk.DragAction.MOVE, time)
                        # Change target as well
                        self.tree.enable_model_drag_dest(
                            [self.playlist_target], Gdk.DragAction.DEFAULT
                        )
                else:
                    Gdk.drag_status(context, Gdk.DragAction.COPY, time)
            elif isinstance(drop_target, TrackWrapper):
                # We are dragging onto another track
                # make it a move operation if we are only dragging
                # tracks within our widget
                # We do a copy if we are draggin from another playlist
                if Gtk.drag_get_source_widget(context) == tv and not dragging_playlist:
                    Gdk.drag_status(context, Gdk.DragAction.MOVE, time)
                else:
                    Gdk.drag_status(context, Gdk.DragAction.COPY, time)
            else:
                # Prevent drop operation by changing the targets
                self.tree.enable_model_drag_dest(
                    self.deny_targets, Gdk.DragAction.DEFAULT
                )
                return False
            return True
        else:  # No drop info
            if dragging_playlist:
                Gdk.drag_status(context, Gdk.DragAction.MOVE, time)
                # Change target as well
                self.tree.enable_model_drag_dest(
                    [self.playlist_target], Gdk.DragAction.DEFAULT
                )
                return True
            return False

    def on_key_released(self, widget, event):
        """
        Called when a key is released in the tree
        """
        if event.keyval == Gdk.KEY_Menu:
            mods, paths = self.tree.get_selection().get_selected_rows()
            if paths and paths[0]:
                iter = self.model.get_iter(paths[0])
                pl = self.model.get_value(iter, 2)
                # Based on what is selected determines what
                # menu we will show
                if isinstance(pl, Playlist):
                    Gtk.Menu.popup(
                        self.playlist_menu, None, None, None, None, 0, event.time
                    )
                elif isinstance(pl, SmartPlaylist):
                    Gtk.Menu.popup(
                        self.smart_menu, None, None, None, None, 0, event.time
                    )
                elif isinstance(pl, TrackWrapper):
                    Gtk.Menu.popup(
                        self.track_menu, None, None, None, None, 0, event.time
                    )
                else:
                    Gtk.Menu.popup(
                        self.default_menu, None, None, None, None, 0, event.time
                    )
            return True

        if event.keyval == Gdk.KEY_Left:
            mods, paths = self.tree.get_selection().get_selected_rows()
            if paths and paths[0]:
                self.tree.collapse_row(paths[0])
            return True

        if event.keyval == Gdk.KEY_Right:
            mods, paths = self.tree.get_selection().get_selected_rows()
            if paths and paths[0]:
                self.tree.expand_row(paths[0], False)
            return True

        if event.keyval == Gdk.KEY_Delete:
            mods, paths = self.tree.get_selection().get_selected_rows()
            if paths and paths[0]:
                iter = self.model.get_iter(paths[0])
                pl = self.model.get_value(iter, 2)
                # Based on what is selected determines what
                # menu we will show
                if isinstance(pl, (Playlist, SmartPlaylist)):
                    self.remove_playlist(pl)
                elif isinstance(pl, TrackWrapper):
                    self.remove_selected_track()
            return True
        return False

    def _clear_node(self, node):
        """
        Clears a node of all children
        """
        iter = self.model.iter_children(node)
        while True:
            if not iter:
                break
            self.model.remove(iter)
            iter = self.model.iter_children(node)


class PlaylistDragTreeView(DragTreeView):
    """
    Custom DragTreeView to retrieve data from playlists
    """

    def __init__(self, container, receive=True, source=True):
        DragTreeView.__init__(self, container, receive, source)
        self.show_cover_drag_icon = False

    def get_selection_empty(self):
        '''Returns True if there are no selected items'''
        return self.get_selection().count_selected_rows() == 0

    def get_selection_is_computed(self):
        """
        Returns True if selection is a Smart Playlist
        """
        item = self.get_selected_item(raw=True)
        return isinstance(item, SmartPlaylist)

    def get_selected_tracks(self):
        """
        Used by the menu, just basically gets the selected
        playlist and returns the tracks in it
        """
        playlist = self.get_selected_page()

        if playlist is not None:
            return [track for track in playlist]
        else:
            return [self.get_selected_track()]

        return None

    def get_selected_page(self, raw=False):
        """
        Retrieves the currently selected playlist in
        the playlists panel.  If a non-playlist is
        selected it returns None

        @return: the playlist
        """
        item = self.get_selected_item(raw=raw)

        if isinstance(item, (Playlist, SmartPlaylist)):
            return item
        else:
            return None

    def get_selected_track(self):
        item = self.get_selected_item()

        if not item:
            return None

        if isinstance(item, TrackWrapper):
            return item.track
        else:
            return None

    def get_selected_item(self, raw=False):
        model, iter = self.get_selection().get_selected()

        if not iter:
            return None

        item = model.get_value(iter, 2)

        # for smart playlists
        if isinstance(item, SmartPlaylist):
            if raw:
                return item
            try:
                return item.get_playlist(self.container.collection)
            except Exception:
                return None
        if isinstance(item, radio.RadioItem):
            if raw:
                return item
            return item.get_playlist()
        elif isinstance(item, Playlist):
            return item
        elif isinstance(item, TrackWrapper):
            return item
        else:
            return None



# Copyright (C) 2006 Adam Olsen
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 1, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

import logging
import time

from xl import common, covers, event, providers, settings

from . import _ecs as ecs
from . import amazonprefs

logger = logging.getLogger(__name__)

AMAZON = None
USER_AGENT = None


def enable(exaile):
    if exaile.loading:
        event.add_callback(_enable, "exaile_loaded")
    else:
        _enable(None, exaile, None)


def _enable(eventname, exaile, nothing):
    global AMAZON, USER_AGENT
    USER_AGENT = exaile.get_user_agent_string('amazoncovers')
    AMAZON = AmazonCoverSearch()
    providers.register('covers', AMAZON)


def disable(exaile):
    providers.unregister('covers', AMAZON)


def get_preferences_pane():
    return amazonprefs


class AmazonCoverSearch(covers.CoverSearchMethod):
    """
    Searches amazon for an album cover
    """

    name = 'amazon'
    title = 'Amazon'

    def __init__(self):
        self.starttime = 0

    def find_covers(self, track, limit=-1):
        """
        Searches amazon for album covers
        """
        try:
            artist = track.get_tag_raw('artist')[0]
            album = track.get_tag_raw('album')[0]
        except (AttributeError, TypeError):
            return []

        # get the settings for amazon key and secret key
        api_key = settings.get_option('plugin/amazoncovers/api_key', '')
        secret_key = settings.get_option('plugin/amazoncovers/secret_key', '')
        if not api_key or not secret_key:
            logger.warning(
                'Please enter your Amazon API and secret '
                'keys in the Amazon Covers preferences'
            )
            return []

        # wait at least 1 second until the next attempt
        waittime = 1 - (time.time() - self.starttime)
        if waittime > 0:
            time.sleep(waittime)
        self.starttime = time.time()

        search = "%s - %s" % (artist, album)
        try:
            albums = ecs.search_covers(search, api_key, secret_key, USER_AGENT)
        except ecs.AmazonSearchError:
            return []
        return albums

    def get_cover_data(self, url):
        return common.get_url_contents(url, USER_AGENT)



# Copyright (C) 2006 Adam Olsen
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 1, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

import urllib.parse
import urllib.request
import hmac
import hashlib
import base64
import datetime
import re
from xl import common
import logging

logger = logging.getLogger(__name__)


class AmazonSearchError(Exception):
    pass


def generate_timestamp():
    ret = datetime.datetime.utcnow()
    return ret.isoformat() + 'Z'


# make a valid RESTful AWS query, that is signed, from a dictionary
# https://docs.aws.amazon.com/AWSECommerceService/latest/DG/RequestAuthenticationArticle.html
# code by Robert Wallis: SmilingRob@gmail.com, your hourly software contractor


def get_aws_query_string(aws_access_key_id, secret, query_dictionary):
    query_dictionary["AWSAccessKeyId"] = aws_access_key_id
    query_dictionary["Timestamp"] = generate_timestamp()
    query_pairs = sorted(
        map(lambda k, v: (k + "=" + urllib.parse.quote(v)), query_dictionary.items())
    )
    # The Amazon specs require a sorted list of arguments
    query_string = "&".join(query_pairs)
    hm = hmac.new(
        secret,
        "GET\nwebservices.amazon.com\n/onca/xml\n" + query_string,
        hashlib.sha256,
    )
    signature = urllib.parse.quote(base64.b64encode(hm.digest()))
    query_string = "https://webservices.amazon.com/onca/xml?%s&Signature=%s" % (
        query_string,
        signature,
    )
    return query_string


def search_covers(search, api_key, secret_key, user_agent):
    params = {
        'Operation': 'ItemSearch',
        'Keywords': str(search),
        'AssociateTag': 'InvalidTag',  # now required for AWS cover search API
        'Version': '2009-01-06',
        'SearchIndex': 'Music',
        'Service': 'AWSECommerceService',
        'ResponseGroup': 'ItemAttributes,Images',
    }

    query_string = get_aws_query_string(
        str(api_key).strip(), str(secret_key).strip(), params
    )

    headers = {'User-Agent': user_agent}
    req = urllib.request.Request(query_string, None, headers)
    data = urllib.request.urlopen(req).read()

    data = common.get_url_contents(query_string, user_agent)

    # check for an error message
    m = re.search(r'<Message>(.*)</Message>', data, re.DOTALL)
    if m:
        logger.warning('Amazon Covers Search Error: %s', m.group(1))
        raise AmazonSearchError(m.group(1))

    # check for large images
    regex = re.compile(r'<LargeImage><URL>([^<]*)', re.DOTALL)
    items = regex.findall(data)

    return items



#!/usr/bin/env python3

# Copyright (C) 2015, 2017-2018  Johannes Sasongko <sasongko@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


def bash_completion(parser):
    """
    :type parser: argparse.ArgumentParser
    """

    TEMPLATE = '''_%(prog)s() {
    local cur="${COMP_WORDS[COMP_CWORD]}"
    local prev="${COMP_WORDS[COMP_CWORD-1]}"
    case "${prev}" in
%(prev_opts)s
    esac
    case "${cur}" in
%(cur_opts)s
    esac
}
complete -F _%(prog)s %(prog)s'''

    TEMPLATE_OPT = '''        %(opt)s)
%(reply)s
            return 0
            ;;'''

    TEMPLATE_REPLY_W = '''            COMPREPLY=( $(compgen -W "%(opts)s" -- ${cur}) )'''
    TEMPLATE_REPLY_f = '''            COMPREPLY=( $(compgen -f ${cur}) )'''
    TEMPLATE_REPLY_d = '''            COMPREPLY=( $(compgen -d ${cur}) )'''
    TEMPLATE_REPLY_empty = '''            COMPREPLY=()'''

    opt_names = []
    filearg_opt_names = []
    dirarg_opt_names = []
    arg_opt_names = []
    for action in parser._actions:
        names = action.option_strings
        if not names:  # Positional arg
            continue
        opt_names.extend(names)
        if action.metavar == 'LOCATION':
            filearg_opt_names.extend(names)
        elif action.metavar == 'DIRECTORY':
            dirarg_opt_names.extend(names)
        elif action.metavar is not None:
            arg_opt_names.extend(names)

    return TEMPLATE % {
        'prog': 'exaile',
        'prev_opts': '\n'.join([
            TEMPLATE_OPT % {
                'opt': '|'.join(filearg_opt_names),
                'reply': TEMPLATE_REPLY_f,
            },
            TEMPLATE_OPT % {
                'opt': '|'.join(dirarg_opt_names),
                'reply': TEMPLATE_REPLY_d,
            },
            TEMPLATE_OPT % {
                'opt': '|'.join(arg_opt_names),
                'reply': TEMPLATE_REPLY_empty,
            },
        ]),
        'cur_opts': '\n'.join([
            TEMPLATE_OPT % {
                'opt': '-*',
                'reply': TEMPLATE_REPLY_W % {'opts': ' '.join(opt_names)},
            },
            TEMPLATE_OPT % {
                'opt': '*',
                'reply': TEMPLATE_REPLY_f,
            },
        ]),
    }


def fish_completion(parser):
    """
    :type parser: argparse.ArgumentParser
    """

    import shlex

    options = []
    for action in parser._actions:
        names = action.option_strings
        if not names:  # Positional arg
            continue
        option = ['complete -c exaile']
        for name in names:
            assert len(name) >= 2 and name[0] == '-' and name != '--'
            if len(name) == 2:
                option.append('-s ' + shlex.quote(name[1]))
            elif name[1] == '-':
                option.append('-l ' + shlex.quote(name[2:]))
            else:
                option.append('-o ' + shlex.quote(name[1:]))
        if action.metavar in ('LOCATION', 'DIRECTORY'):
            option.append('-r')
        elif action.metavar is not None:
            option.append('-x')
        if action.choices:
            choices = action.choices
            if isinstance(choices, (list, tuple)):
                choices = (shlex.quote(str(c))+'\\t' for c in choices)
                option.append('-a ' + shlex.quote(' '.join(choices)))
        if action.help:
            option.append('-d ' + shlex.quote(action.help % action.__dict__))
        options.append(' '.join(option))

    return '\n'.join(options)


if __name__ == '__main__':
    import os
    import sys
    root = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))
    os.environ['LC_ALL'] = 'C'  # Avoid getting translated metavars
    os.environ['EXAILE_DIR'] = root
    sys.path.insert(1, root)
    from xl import main
    p = main.create_argument_parser()
    if len(sys.argv) < 2 or sys.argv[1] == 'bash':
        completion = bash_completion
    elif sys.argv[1] == 'fish':
        completion = fish_completion
    print(completion(p))



# Filter Widget/Dialog
# Copyright (C) 2006 Johannes Sasongko <sasongko@gmail.com>
# Copyright (C) 2008-2010 Adam Olsen
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
#
# The developers of the Exaile media player hereby grant permission
# for non-GPL compatible GStreamer and Exaile plugins to be used and
# distributed together with GStreamer and Exaile. This permission is
# above and beyond the permissions granted by the GPL license by which
# Exaile is covered. If you modify this code, you may extend this
# exception to your version of the code, but you are not obligated to
# do so. If you do not wish to do so, delete this exception statement
# from your version.

"""Generic widget and dialog for filtering items.

They resemble the configuration dialogs of Evolution's mail filters
and Rhythmbox's automatic playlists.
"""

import urllib.parse
from typing import Sequence

from gi.types import GObjectMeta
from gi.repository import Gtk

from xl.nls import gettext as _
from xlgui.guiutil import GtkTemplate
from xlgui.guiutil import gtk_widget_replace


@GtkTemplate('ui', 'widgets', 'filter_dialog.ui')
class FilterDialog(Gtk.Dialog):
    """Dialog to filter a list of items.

    Consists of a FilterWidget and an Add button.
    """

    __gtype_name__ = 'FilterDialog'

    (
        name_entry,
        filter,
        match_any,
        random,
        lim_check,
        lim_spin,
    ) = GtkTemplate.Child.widgets(6)

    def __init__(self, title, parent, criteria):
        """Create a filter dialog.

        Parameters:
        - title: title of the dialog window
        - criteria: possible criteria; see FilterWindow
        """

        Gtk.Dialog.__init__(self, title=title, transient_for=parent)
        self.init_template()

        self.add_buttons(
            Gtk.STOCK_CANCEL,
            Gtk.ResponseType.REJECT,
            Gtk.STOCK_OK,
            Gtk.ResponseType.ACCEPT,
        )

        f = FilterWidget(sorted(criteria, key=lambda k: _(k[0])))
        f.add_criteria_row()
        f.set_border_width(5)
        f.show_all()

        self.filter = gtk_widget_replace(self.filter, f)

    @GtkTemplate.Callback
    def on_add_button_clicked(self, *args):
        self.filter.add_criteria_row()

    @GtkTemplate.Callback
    def on_lim_check_toggled(self, *args):
        self.lim_spin.set_sensitive(self.lim_check.get_active())

    def set_limit(self, limit):
        """
        Sets the limit for the number of items that should be returned
        """
        if limit > -1:
            self.lim_check.set_active(True)
            self.lim_spin.set_value(limit)
        else:
            self.lim_check.set_active(False)

    def get_limit(self):
        """
        Get the limit value
        """
        if self.lim_check.get_active():
            return int(self.lim_spin.get_value())
        else:
            return -1

    def get_random(self):
        """
        Returns if the playlist should be random
        """
        return self.random.get_active()

    def set_random(self, random):
        """
        Sets if this playlist should be random
        """
        self.random.set_active(random)

    def get_match_any(self):
        """
        Returns true if this dialog should match any of the criteria
        """
        return self.match_any.get_active()

    def set_match_any(self, any):
        """
        Sets whether this dialog should match any of the criteria or not
        """
        self.match_any.set_active(any)

    def get_name(self):
        """
        Returns the text in the name_entry
        """
        return self.name_entry.get_text()

    def set_name(self, name):
        """
        Sets the text in the name_entry
        """
        self.name_entry.set_text(name)

    def get_result(self):
        """Return the user's input as a list of filter criteria.

        See FilterWidget.get_result.
        """
        # TODO: display error message
        return self.filter.get_result()

    def get_state(self):
        """Return the filter state.

        See FilterWidget.get_state.
        """
        return self.filter.get_state()

    def set_state(self, state):
        """Set the filter state.

        See FilterWidget.set_state.
        """
        self.filter.set_state(state)


class FilterWidget(Gtk.Grid):
    """Widget to filter a list of items.

    This widget only includes the criteria selector (Criterion widgets)
    and their Remove buttons; it does not include an Add button.

    Attributes:
    - criteria: see example
    - rows: list of (criterion, remove_btn, remove_btn_handler_id)
    - n: number of rows

    Example of criteria:

        [
            # name
            (N_('Year'), [
                # name - field class/factory
                (N_('is'), EntryField),
                (N_('is between'),
                    lambda x: EntryLabelEntryField(x, _('and'))),
                (N_('is this year'), NothingField)
            ]),
        ]

    The field class is a class following this interface:

        class Field(Gtk.Widget):
            def __init__(self, result_generator):
                pass
            def get_result(self):
                return object
            def get_state(self):
                return object
            def set_state(self, state):
                pass

    Although not required, these methods should use unicode instead of
    str objects when dealing with strings.
    """

    def __init__(self, criteria):
        """Create a filter widget.

        Parameter:
        - criteria: see FilterWidget
        """

        super(FilterWidget, self).__init__()
        self.set_column_spacing(10)
        self.set_row_spacing(2)
        self.criteria = criteria
        self.rows = []

    def add_criteria_row(self):
        """Add a new criteria row."""
        criterion = Criterion(self.criteria)
        criterion.show()

        n = len(self.rows)

        if n != 0:
            criterion.set_state(self.rows[-1][0].get_state())

        remove_btn = Gtk.Button()
        image = Gtk.Image()
        image.set_from_icon_name('list-remove', Gtk.IconSize.BUTTON)
        remove_btn.add(image)
        remove_btn_handler_id = remove_btn.connect('clicked', self.__remove_clicked)
        remove_btn.show_all()

        self.attach(criterion, 0, n, 1, 1)
        self.attach(remove_btn, 1, n, 1, 1)

        self.rows.append((criterion, remove_btn, remove_btn_handler_id))

    def remove_criteria_row(self, row):
        """Remove a criteria row."""
        self.remove_row(row)
        del self.rows[row]

    def __remove_clicked(self, widget):
        n = self.child_get_property(widget, 'top-attach')
        self.remove_criteria_row(n)

    def get_state(self):
        """Return the filter state.

        See set_state for the state format.
        """
        state = []
        for row in self.rows:
            state.append(row[0].get_state())
            state[-1][0].reverse()  # reverse so it reads more nicely
        return state

    def set_state(self, state):
        """Set the filter state.

        Format:

            [
                ( [criterion1_1, criterion1_2, ...], filter1 ),
                ( [criterion2_1, criterion2_2, ...], filter2 ),
                ...
            ]
        """
        n_present = len(self.rows)
        n_required = len(state)
        for i in range(n_present, n_required):
            self.add_criteria_row()
        for i in range(n_present, n_required, -1):
            self.remove_criteria_row(i - 1)  # i is one less than n
        for i, cstate in enumerate(state):
            cstate[0].reverse()  # reverse so it becomes a stack
            self.rows[i][0].set_state(cstate)


class Criterion(Gtk.Box):
    """Widget representing one filter criterion.

    It contains either:
    - one combo box and another criterion object, or
    - a field object.
    """

    def __init__(self, subcriteria):
        """Create a criterion object.

        Parameters:
        - subcriteria: a list of possible subcriteria;
          see criteria in FilterWidget
        """
        super(Criterion, self).__init__(spacing=5)
        if isinstance(subcriteria, (Gtk.Widget, GObjectMeta)):
            field_class = subcriteria
            self.child = field_class()
        else:
            self.combo = combo = Gtk.ComboBoxText()
            if len(subcriteria) > 10:
                self.combo.set_wrap_width(5)
            self.subcriteria = subcriteria
            for subc in subcriteria:
                combo.append_text(_(subc[0]))
            combo.set_active(0)
            combo.connect('changed', self._combo_changed)
            combo.show()
            self.pack_start(combo, False, True, 0)
            self.child = Criterion(subcriteria[0][1])
        self.child.show()
        self.pack_start(self.child, True, True, 0)

    def _combo_changed(self, widget):
        """Called when the combo box changes its value."""
        state = self.child.get_state()
        self.remove(self.child)
        self.child = Criterion(self.subcriteria[self.combo.get_active()][1])
        if state:
            self.child.set_state(state)
        self.pack_start(self.child, True, True, 0)
        self.child.show()

    def get_state(self):
        """Return the criterion state.

        See set_state for the state format.
        """
        state = self.child.get_state()
        if isinstance(self.child, Criterion):
            state[0].append(self.subcriteria[self.combo.get_active()][0])
        else:
            state = ([], state)
        return state

    def set_state(self, state):
        """Set the criterion state.

        Format:

            ([..., grandchild_state, child_state, self_state], filter)

        Note the reverse order of the list. This is to give the
        impression of it being a stack responding to pop().
        """
        if isinstance(self.child, Criterion):
            text = state[0].pop()
            for i, subc in enumerate(self.subcriteria):
                if subc[0] == text:
                    self.combo.set_active(i)
                    break
            self.child.set_state(state)
        else:
            if len(state) > 1:
                self.child.set_state(state[1])


# Sample fields


class ComboEntryField(Gtk.Box):
    '''Select from multiple fixed values, but allow the user to enter text'''

    def __init__(self, values):
        Gtk.Box.__init__(self)

        self.combo = Gtk.ComboBoxText.new_with_entry()
        for value in values:
            self.combo.append_text(value)

        self.pack_start(self.combo, True, True, 0)
        self.combo.show()

    def get_state(self):
        return self.combo.get_active_text()

    def set_state(self, state):
        self.combo.get_child().set_text(state)


class NullField(Gtk.Box):
    '''Used as a placeholder for __null__ values'''

    def get_state(self):
        return ['__null__']

    def set_state(self, state):
        pass


class MultiEntryField(Gtk.Box):
    """Helper field that can be subclassed to get fields with multiple
    GtkEntry widgets and multiple labels."""

    def __init__(self, labels):
        """Create a field with the specified labels and widths.

        Parameter:
        - labels: sequence of string, integer, or None values;
          string represents label,
          integer represents Entry widget with a specific width,
          None represents Entry widget with default width
        """
        Gtk.Box.__init__(self, spacing=5)
        self.entries = []
        for label in labels:
            if label is None:
                widget = Gtk.Entry()
                self.entries.append(widget)
            elif isinstance(label, (int, float)):
                widget = Gtk.Entry()
                widget.set_size_request(label, -1)
                self.entries.append(widget)
            else:
                widget = Gtk.Label(label=str(label))
            self.pack_start(widget, False, True, 0)
            widget.show()

    def get_state(self):
        return [e.get_text() for e in self.entries]

    def set_state(self, state):
        entries = self.entries
        if isinstance(state, (list, tuple)):
            for i in range(min(len(entries), len(state))):
                entries[i].set_text(state[i])
        else:
            entries[0].set_text(state)


class EntryField(Gtk.Entry):
    def __init__(self):
        Gtk.Entry.__init__(self)

    def get_state(self):
        return self.get_text()

    def set_state(self, state):
        if isinstance(state, list) or isinstance(state, tuple):
            state = state[0]
        self.set_text(state)


class QuotedEntryField(Gtk.Entry):
    def __init__(self):
        Gtk.Entry.__init__(self)

    def get_state(self):
        return urllib.parse.quote(self.get_text())

    def set_state(self, state):
        if isinstance(state, list) or isinstance(state, tuple):
            state = state[0]
        self.set_text(urllib.parse.unquote(state))


class EntryLabelEntryField(MultiEntryField):
    def __init__(self, label):
        MultiEntryField.__init__(self, (50, label, 50))


class SpinLabelField(Gtk.Box):
    def __init__(self, label='', top=999999, lower=-999999):
        Gtk.Box.__init__(self, spacing=5)
        self.spin = Gtk.SpinButton.new_with_range(lower, top, 1)
        self.spin.set_value(0)
        self.pack_start(self.spin, False, True, 0)
        self.pack_start(Gtk.Label.new(label), False, True, 0)
        self.show_all()

    def get_state(self):
        return self.spin.get_value()

    def set_state(self, state):
        if isinstance(state, list) or isinstance(state, tuple):
            state = state[0]
        try:
            self.spin.set_value(int(state))
        except ValueError:
            pass


class SpinButtonAndComboField(Gtk.Box):
    def __init__(self, items=()):
        Gtk.Box.__init__(self, spacing=5)
        self.items: Sequence[str] = items

        self.entry = Gtk.SpinButton.new_with_range(0, 999999, 1)
        self.entry.set_value(0)
        self.pack_start(self.entry, True, True, 0)

        self.combo = Gtk.ComboBoxText()
        for item in items:
            self.combo.append_text(_(item))
        self.combo.set_active(0)
        self.pack_start(self.combo, True, True, 0)
        self.show_all()

    def set_state(self, state):
        if not isinstance(state, (tuple, list)) or len(state) < 2:
            return

        # TODO: Check length.
        try:
            self.entry.set_value(int(state[0]))
        except ValueError:
            pass
        combo_state = state[1]
        try:
            index = self.items.index(combo_state)
        except ValueError:
            pass
        else:
            self.combo.set_active(index)

    def get_state(self):
        active_item = self.items[self.combo.get_active()]
        return [self.entry.get_value(), active_item]



# -*- coding: utf-8
# Copyright (C) 2008-2010 Adam Olsen
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
#
# The developers of the Exaile media player hereby grant permission
# for non-GPL compatible GStreamer and Exaile plugins to be used and
# distributed together with GStreamer and Exaile. This permission is
# above and beyond the permissions granted by the GPL license by which
# Exaile is covered. If you modify this code, you may extend this
# exception to your version of the code, but you are not obligated to
# do so. If you do not wish to do so, delete this exception statement
# from your version.

from copy import deepcopy
import logging
import operator
import re
import time
from typing import Dict, Generic, List, Optional, TypeVar, Union
import unicodedata
import weakref

from gi.repository import Gio
from gi.repository import GLib

from xl.metadata._base import BaseFormat
import xl.unicode
from xl import event, metadata, settings
from xl.metadata.tags import disk_tags
from xl.nls import gettext as _
from xl.unicode import shave_marks

logger = logging.getLogger(__name__)

_K = TypeVar('_K')
_V = TypeVar('_V')
_V1 = TypeVar('_V1')

# map chars to appropriate substitutes for sorting
_sortcharmap = {
    'ß': 'ss',  # U+00DF
    'æ': 'ae',  # U+00E6
    'ĳ': 'ij',  # U+0133
    'ŋ': 'ng',  # U+014B
    'œ': 'oe',  # U+0153
    'ƕ': 'hv',  # U+0195
    'ǆ': 'dz',  # U+01C6
    'ǉ': 'lj',  # U+01C9
    'ǌ': 'nj',  # U+01CC
    'ǳ': 'dz',  # U+01F3
    'ҥ': 'ng',  # U+04A5
    'ҵ': 'ts',  # U+04B5
}

# Cache these here because calling gettext inside get_tag_display
# is two orders of magnitude slower.
_VARIOUSARTISTSSTR = _("Various Artists")
_UNKNOWNSTR = _("Unknown")
# TRANSLATORS: String multiple tag values will be joined by
_JOINSTR = _(' / ')

_no_set_raw = {'__basename', '__loc'} | disk_tags

_unset = object()


class _MetadataCacher(Generic[_K, _V]):
    """Time- and size-limited LRU cache"""

    class Entry(Generic[_V1]):
        value: _V1
        time: float

        def __init__(self, value: _V1, time: float):
            self.value = value
            self.time = time

        def __lt__(self, other: '_MetadataCacher.Entry'):
            return self.time < other.time

    _cache: Dict[_K, Entry[_V]]
    timeout: int
    maxentries: int
    _cleanup_id: Optional[int]

    def __init__(self, timeout: int = 10, maxentries: int = 20):
        """
        :param timeout: time (in s) until the cached obj gets removed.
        :param maxentries: maximum number of objects to cache
        """
        self._cache = {}
        self.timeout = timeout
        self.maxentries = maxentries
        self._cleanup_id = None

    def __cleanup(self) -> bool:
        if self._cleanup_id is not None:
            GLib.source_remove(self._cleanup_id)
            self._cleanup_id = None
        current = time.time()
        thresh = current - self.timeout
        self._cache = {k: v for k, v in self._cache.items() if v.time >= thresh}
        if self._cache:
            next_expiry = min(self._cache.values()).time
            timeout = int((next_expiry + self.timeout) - current)
            self._cleanup_id = GLib.timeout_add_seconds(timeout, self.__cleanup)
        return False

    def add(self, key: _K, value: _V) -> None:
        if key in self._cache:
            return
        item = self.Entry(value, time.time())
        self._cache[key] = item
        if len(self._cache) > self.maxentries:
            oldest_key = min(self._cache.items(), key=operator.itemgetter(1))[0]
            del self._cache[oldest_key]
        if self._cleanup_id is None:
            self._cleanup_id = GLib.timeout_add_seconds(self.timeout, self.__cleanup)

    def remove(self, key: _K) -> None:
        try:
            del self._cache[key]
        except KeyError:
            pass

    def get(self, key: _K) -> Optional[_V]:
        item = self._cache.get(key)
        if item is None:
            return None
        item.time = time.time()
        return item.value


#: Cache of metadata format objects to speed up get_tag_disk
_CACHER: _MetadataCacher['Track', BaseFormat] = _MetadataCacher()


class Track:
    """
    Represents a single track.
    """

    # save a little memory this way
    __slots__ = [
        "__tags",
        "_scan_valid",
        "_dirty",
        "__weakref__",
        "_init",
        "_is_supported",
    ]
    # this is used to enforce the one-track-per-uri rule
    __tracksdict = weakref.WeakValueDictionary()
    # store a copy of the settings values here - much faster (0.25 cpu
    # seconds) (see _the_cuts_cb)
    __the_cuts = settings.get_option('collection/strip_list', [])

    def __new__(cls, *args, **kwargs):
        """
        override the construction of new Track objects so that
        if there is already a Track for a given uri, we just return
        that Track instance instead of creating a new one.
        """
        # subclassing interferes with the one-track-per-uri scheme and
        # with save and restore of tracks, so we disallow it.
        if cls != Track:
            raise TypeError("Track cannot be subclassed!")

        uri = None
        if len(args) > 0:
            uri = args[0]
        else:
            uri = kwargs.get("uri")

        # Restore uri from pickled state if possible.  This means that
        # if a given Track is in more than one TrackDB, the first
        # TrackDB to get loaded takes precedence, and any data in the
        # second TrackDB is consequently ignored. Thus if at all
        # possible, Tracks should NOT be persisted in more than one
        # TrackDB at a time.
        unpickles = None
        if uri is None:
            if len(args) > 2:
                unpickles = args[2]
            else:
                unpickles = kwargs.get("_unpickles")
            if unpickles is not None:
                uri = unpickles.get("__loc")

        if uri is not None:
            uri = Gio.File.new_for_uri(uri).get_uri()
            try:
                tr = cls.__tracksdict[uri]
                tr._init = False

                # if the track *does* happen to be pickled in more than one
                # place, then we need to preserve any internal tags that aren't
                # persisted to disk.
                #
                # See https://bugs.launchpad.net/exaile/+bug/1054637
                if unpickles is None:
                    if len(args) > 2:
                        unpickles = args[2]
                    else:
                        unpickles = kwargs.get("_unpickles")

                if unpickles is not None:
                    tags = tr.list_tags()
                    to_set = {
                        tag: values
                        for tag, values in unpickles.items()
                        if tag.startswith('__') and tag not in tags
                    }
                    if to_set:
                        tr.set_tags(**to_set)

            except KeyError:
                tr = object.__new__(cls)
                cls.__tracksdict[uri] = tr
                tr._init = True
            return tr
        else:
            # this should always fail in __init__, and will never be
            # called in well-formed code.
            tr = object.__new__(cls)
            tr._init = True
            return tr

    def __init__(self, uri: Optional[str] = None, scan: bool = True, _unpickles=None):
        """
        :param uri: the location, as either a uri or a file path.
        :param scan: Whether to try to read tags from the given uri.
              Use only if the tags need to be set by a
              different source.

        :param _unpickles: used internally to restore from a pickled
            state. not for normal use.
        """
        # don't re-init if its a reused track. see __new__
        if self._init is False:
            return

        self.__tags = {}
        self._scan_valid = None  # whether our last tag read attempt worked
        self._is_supported = None

        # This is not used by write_tags, this is used by the collection to
        # indicate that the tags haven't been written to the collection
        self._dirty = False

        if _unpickles:
            self._unpickles(_unpickles)
            self.__register()
        elif uri:
            # notify isn't needed here because this is a new track
            self.set_loc(uri, notify_changed=False)
            if scan:
                self.read_tags(notify_changed=False)
        else:
            raise ValueError("Cannot create a Track from nothing")

    def __register(self):
        """
        Register this instance into the global registry of Track
        objects.
        """
        self.__tracksdict[self.__tags['__loc']] = self

    def __unregister(self):
        """
        Unregister this instance from the global registry of
        Track objects.
        """
        try:
            del self.__tracksdict[self.__tags['__loc']]
        except KeyError:
            pass

    def set_loc(self, loc: str, notify_changed: bool = True) -> None:
        """
        Sets the location.

        :param loc: the location, as either a uri or a file path.
        """
        self.__unregister()
        gloc = Gio.File.new_for_commandline_arg(loc)
        self.__tags['__loc'] = gloc.get_uri()
        self.__register()
        if notify_changed:
            event.log_event('track_tags_changed', self, {'__loc'})

    def exists(self):
        """
        Returns whether the file exists
        This can be very slow, use with caution!
        """
        return Gio.File.new_for_uri(self.get_loc_for_io()).query_exists(None)

    def get_loc_for_io(self) -> str:
        """
        Gets the location as a full uri.

        Safe for IO operations via gio, not suitable for display to users
        as it may be in non-utf-8 encodings.
        """
        return self.__tags['__loc']

    def get_local_path(self):
        """
        If the file is accessible on a local filesystem, retrieves
        the full path to it, otherwise nothing.

        :returns: the file path or None
        :rtype: string or NoneType
        """
        return Gio.File.new_for_uri(self.get_loc_for_io()).get_path()

    def get_basename(self):
        """
        Returns the base name of a resource
        """
        gfile = Gio.File.new_for_uri(self.get_loc_for_io())

        return gfile.get_basename()

    def get_basename_display(self) -> str:
        """
        Return the track basename for display purposes (invalid characters
        are replaced).

        :rtype: unicode
        """
        gfile = Gio.File.new_for_uri(self.get_loc_for_io())
        path = gfile.get_path()
        if path:  # Local
            path = GLib.filename_display_basename(path)
        else:  # Non-local
            path = GLib.filename_display_name(gfile.get_basename())
        return path

    def get_type(self):
        """
        Get the URI schema the file uses, e.g. file, http, smb.
        """
        return Gio.File.new_for_uri(self.get_loc_for_io()).get_uri_scheme()

    def write_tags(self):
        """
        Writes tags to the file for this Track.

        Returns False if unsuccessful, and a Format object from
        `xl.metadata` otherwise.
        """
        try:
            f = metadata.get_format(self.get_loc_for_io())
            if f is None:
                return False  # not a supported type
            f.write_tags(self.__tags)

            # now that we've written the tags to disk, remove any tags that the
            # user asked to be deleted
            to_remove = [k for k, v in self.__tags.items() if v is None]
            for rm in to_remove:
                self.__tags.pop(rm)

            return f
        except IOError:
            # error writing to the file, probably
            logger.warning("Could not write tags to file", exc_info=True)
            return False
        except Exception:
            logger.exception("Unknown exception: Could not write tags to file")
            return False

    def read_tags(self, force=True, notify_changed=True):
        """
        Reads tags from the file for this Track.

        :param force: If not True, then only read the tags if the file has
                      be modified.

        Returns False if unsuccessful, and a Format object from
        `xl.metadata` otherwise.
        """

        if not self.is_supported():
            self._scan_valid = False
            return False

        loc = self.get_loc_for_io()
        try:
            # Retrieve file specific metadata
            gloc = Gio.File.new_for_uri(loc)
            mtime = (
                gloc.query_info("time::modified", Gio.FileQueryInfoFlags.NONE, None)
                .get_modification_date_time()
                .to_unix()
            )
            f = metadata.get_format(loc)
            if not force and self.__tags.get('__modified', 0) >= mtime:
                return f

            # Read the tags
            ntags = f.read_all()
            ntags['__modified'] = mtime

            # TODO: this probably breaks on non-local files
            ntags['__basedir'] = gloc.get_parent().get_path()

            if '__rating' in ntags and settings.get_option(
                'collection/write_rating_to_audio_file_metadata', False
            ):
                ntags['__rating'] = int(ntags['__rating'][0])

            # remove tags that could be in the file, but are in fact not
            # in the file. Retain tags in the DB that aren't supported by
            # the file format.

            nkeys = set(ntags.keys())
            ekeys = {k for k in self.__tags if not k.startswith('__')}

            # delete anything that wasn't in the new tags
            to_del = ekeys - nkeys

            # but if not others set, only delete supported tags
            if not f.others:
                to_del &= set(f.tag_mapping.keys())

            for tag in to_del:
                ntags[tag] = None

            self.set_tags(notify_changed=notify_changed, **ntags)

            self._scan_valid = True
            return f
        except Exception:
            self._scan_valid = False
            logger.exception("Error reading tags for %s", loc)
            return False

    def is_local(self):
        """
        Determines whether a file is accessible on the local filesystem.
        """
        # TODO: determine this better
        # Maybe use Gio.File.is_native()?
        if self.get_local_path():
            return True
        return False

    def is_supported(self):
        """
        Determines if a file has a supported media format
        """
        if self._is_supported is None:
            loc = self.get_loc_for_io()
            f = metadata.get_format(loc)
            if f is None:
                self._is_supported = False
            else:
                self._is_supported = True

        return self._is_supported

    def get_size(self):
        """
        Get the raw size of the file. Potentially slow.
        """
        f = Gio.File.new_for_uri(self.get_loc_for_io())
        return f.query_info(
            "standard::size", Gio.FileQueryInfoFlags.NONE, None
        ).get_size()

    def __repr__(self):
        return str(self)

    def __str__(self):
        """
        returns a string representing the track
        """
        vals = map(self.get_tag_display, ('title', 'artist', 'album'))
        return "<Track %r by %r from %r>" % tuple(vals)

    def _pickles(self):
        """
        returns a data repr of the track suitable for pickling

        internal use only please
        """
        return deepcopy(self.__tags)

    def _unpickles(self, pickle_obj):
        """
        restores the state from the pickle-able repr

        internal use only please
        """
        self.__tags = deepcopy(pickle_obj)

    def list_tags(self):
        """
        Returns a list of the names of all tags present in this Track.
        """
        return [k for k, v in self.__tags.items() if v is not None] + ['__basename']

    def _xform_set_values(self, tag, values):
        # Handle values that aren't lists
        if not isinstance(values, list):
            if not tag.startswith("__"):  # internal tags dont have to be lists
                values = [values]

        if (
            tag == '__rating'
            and not isinstance(values, list)
            and self._write_rating_to_disk()
        ):
            values = [values]

        # For lists, filter out empty values
        if isinstance(values, list):
            values = [v for v in values if v not in (None, '')]

        if values:
            return values

        return None

    def set_tag_raw(self, tag, values, notify_changed=True):
        """
        Set the raw value of a tag.

        :param tag: The name of the tag to set.
        :param values: The value or values to set the tag to.
        :param notify_changed: whether to send a signal to let other
            parts of Exaile know there has been an update. Only set
            this to False if you know that no other parts of Exaile
            need to be updated.

        .. note:: When setting more than one tag, prefer set_tags instead

        .. warning:: Covers and lyrics tags must be set via set_tag_disk

        :returns: True if changed, False otherwise
        """
        changed = self.set_tags(notify_changed=notify_changed, **{tag: values})
        return bool(changed)

    def set_tags(self, notify_changed=True, **kwargs):
        """
        Set multiple tags on a track.

        :param notify_changed: whether to send a signal to let other
            parts of Exaile know there has been an update. Only set
            this to False if you know that no other parts of Exaile
            need to be updated.

        Prefer this method over calling set_tag_raw multiple times, as this
        method will be more efficient.

        .. warning:: Covers and lyrics tags must be set via set_tag_disk

        :returns: Set of tags that have changed
        """

        # tag changes can cause expensive UI updates, so don't emit the event
        # if the track hasn't actually changed
        changed = set()

        for tag, values in kwargs.items():
            if tag in _no_set_raw:
                if tag == '__loc':
                    logger.warning(
                        'Setting "__loc" directly is forbidden, use set_loc() instead.'
                    )
                elif tag == '__basename':
                    logger.warning('Setting "__basename" directly is forbidden.')
                else:
                    logger.warning(
                        'Cannot set "%s" via set_tag_raw, use set_tag_disk instead', tag
                    )
                continue

            # Transform and set the value. We do NOT delete the value from the tag
            # dict (which was done prior to Exaile 4), otherwise we don't know that
            # the user wanted the tag to be deleted
            new_value = self._xform_set_values(tag, values)
            if self.__tags.get(tag, _unset) != new_value:
                changed.add(tag)
                self.__tags[tag] = new_value

        if changed:
            self._dirty = True
            if notify_changed:
                event.log_event("track_tags_changed", self, changed)

        return changed

    def get_tag_raw(self, tag, join=False):
        """
        Get the raw value of a tag.  For non-internal tags, the
        result will always be a list of unicode strings.

        :param tag: The name of the tag to get
        :param join: If True, joins lists of values into a
            single value.

        :returns: None if the tag is not present
        """
        if tag == '__basename':
            value = self.get_basename()
        elif tag == '__startoffset':
            # TODO: This is only necessary because some places that deal with
            # __startoffset don't check for None. Those need to be fixed.
            value = self.__tags.get(tag, 0)
        elif tag == '__rating':
            try:
                value = self.__tags.get(tag)[0]
            except TypeError as e:
                value = self.__tags.get(tag)
        else:
            value = self.__tags.get(tag)

        if join and value and not tag.startswith('__'):
            return self.join_values(value)

        return value

    def get_tag_sort(
        self, tag, join=True, artist_compilations=False, extend_title=True
    ):
        """
        Get a tag value in a form suitable for sorting.

        :param tag: The name of the tag to get
        :param join: If True, joins lists of values into a
            single value.
        :param artist_compilations: If True, automatically handle
            albumartist and other compilations detections when
            tag=="albumartist".
        :param extend_title: If the title tag is unknown, try to
            add some identifying information to it.
        """
        # The two magic values here are to ensure that compilations
        # and unknown values are always sorted below all normal
        # values.
        value = None
        sorttag = self.__tags.get(tag + "sort")
        if sorttag and tag != "albumartist":
            value = sorttag
        elif tag == "albumartist":
            if artist_compilations and self.__tags.get('__compilation'):
                value = self.__tags.get('albumartist', "\uffff\uffff\uffff\ufffe")
            else:
                value = self.__tags.get('albumartist')
                if value is None:
                    value = self.__tags.get('artist', "\uffff\uffff\uffff\uffff")
            if sorttag and value not in (
                "\uffff\uffff\uffff\ufffe",
                "\uffff\uffff\uffff\uffff",
            ):
                value = sorttag
            else:
                sorttag = None
        elif tag in ('tracknumber', 'discnumber'):
            value = self.split_numerical(self.__tags.get(tag))[0] or 0
        elif tag in ('__length', '__playcount'):
            value = self.__tags.get(tag, 0)
        elif tag == 'bpm':
            try:
                value = int(self.__tags.get(tag, [0])[0])
            except ValueError:
                digits = re.search(r'\d+\.?\d*', self.__tags.get(tag, [0])[0])
                if digits:
                    value = float(digits.group())
        elif tag == '__basename':
            # TODO: Check if unicode() is required
            value = self.get_basename()
        elif tag == '__rating':
            value = self.get_rating()
        else:
            value = self.__tags.get(tag)

        if value is None:
            value = "\uffff\uffff\uffff\uffff"  # unknown
            if tag == 'title':
                basename = self.get_basename_display()
                value = "%s (%s)" % (value, basename)
        elif not tag.startswith("__") and tag not in (
            'tracknumber',
            'discnumber',
            'bpm',
        ):
            if not sorttag:
                value = self.format_sort(value)
            else:
                if isinstance(value, list):
                    value = [self.lower(v + " " + v) for v in value]
                else:
                    value = self.lower(value + " " + value)
            if join:
                value = self.join_values(value)

        return value

    def get_tag_display(
        self, tag, join=True, artist_compilations=False, extend_title=True
    ) -> Union[str, List[str]]:
        """
        Get a tag value in a form suitable for display.

        :param tag: The name of the tag to get
        :param join: If True, joins lists of values into a
            single value.
        :param artist_compilations: If True, automatically handle
            albumartist and other compilations detections when
            tag=="albumartist".
        :param extend_title: If the title tag is unknown, try to
            add some identifying information to it.
        """
        if tag == '__loc':
            return Gio.File.new_for_uri(self.__tags['__loc']).get_parse_name()

        value = None
        if tag == "albumartist":
            if artist_compilations and self.__tags.get('__compilation'):
                value = self.__tags.get('albumartist', _VARIOUSARTISTSSTR)
            else:
                value = self.__tags.get('albumartist')
                if value is None:
                    value = self.__tags.get('artist', _UNKNOWNSTR)
        elif tag in ('tracknumber', 'discnumber'):
            value = self.split_numerical(self.__tags.get(tag))[0] or ""
        elif tag in ('__length', '__startoffset', '__stopoffset'):
            value = self.__tags.get(tag, "")
        elif tag in ('__rating', '__playcount'):
            value = self.__tags.get(tag, "0")
        elif tag == '__bitrate':
            try:
                value = int(self.__tags['__bitrate']) // 1000
                if value == -1:
                    value = ""
                else:
                    # TRANSLATORS: Bitrate (k here is short for kbps).
                    value = _("%dk") % value
            except (KeyError, ValueError):
                value = ""
        elif tag == '__basename':
            value = self.get_basename_display()
        else:
            value = self.__tags.get(tag)

        if value is None:
            value = ''
            if tag == 'title':
                basename = self.get_basename_display()
                value = "%s (%s)" % (_UNKNOWNSTR, basename)

        # Convert value to str or List[str]
        if isinstance(value, list):
            value = [xl.unicode.to_unicode(x, errors='replace') for x in value]
        else:
            value = xl.unicode.to_unicode(value, errors='replace')

        if join:
            value = self.join_values(value, _JOINSTR)

        return value

    def get_tag_search(
        self, tag, format=True, artist_compilations=False, extend_title=True
    ):
        """
        Get a tag value suitable for passing to the search system.
        This includes quoting and list joining.

        :param format: pre-format into a search query.
        :param artist_compilations: If True, automatically handle
            albumartist and other compilations detections when
            tag=="albumartist".
        :param extend_title: If the title tag is unknown, try to
            add some identifying information to it.

        :returns: unicode string that is used for searching
        """
        extraformat = ""
        if tag == "albumartist":
            if artist_compilations and self.__tags.get('__compilation'):
                value = self.__tags.get('albumartist', None)
                tag = 'albumartist'
                extraformat += " ! __compilation==__null__"
            else:
                value = self.__tags.get('albumartist')
                if value is None:
                    value = self.__tags.get('artist')
        elif tag in ('tracknumber', 'discnumber'):
            value = self.split_numerical(self.__tags.get(tag))[0]
        elif tag in (
            '__length',
            '__playcount',
            '__rating',
            '__startoffset',
            '__stopoffset',
        ):
            value = self.__tags.get(tag, 0)
        elif tag == '__bitrate':
            try:
                value = int(self.__tags['__bitrate']) // 1000
                if value != -1:
                    # TRANSLATORS: Bitrate (k here is short for kbps).
                    value = [_("%dk") % value, self.__tags['__bitrate']]
            except (KeyError, ValueError):
                value = -1
        elif tag == '__basename':
            value = self.get_basename_display()
        else:
            value = self.__tags.get(tag)

        # Quote arguments
        if value is None:
            value = '__null__'
            if tag == 'title':
                extraformat += ' __loc==\"%s\"' % self.__tags['__loc']
        elif format:
            if isinstance(value, list):
                value = ['"%s"' % self.quoter(val) for val in value]
            else:
                value = '"%s"' % self.quoter(value)

        # Join lists
        if format:
            if isinstance(value, list):
                value = " ".join(['%s==%s' % (tag, v) for v in value])
            else:
                value = '%s==%s' % (tag, value)
            if extraformat:
                value += extraformat

        # Shave marks from strings
        if isinstance(value, list):
            value = [shave_marks(v) for v in value]
        else:
            value = shave_marks(value)

        return value

    def _get_format_obj(self):
        f = _CACHER.get(self)
        if not f:
            try:
                f = metadata.get_format(self.get_loc_for_io())
            except Exception:  # TODO: What exception?
                return None
            if not f:
                return None
        _CACHER.add(self, f)
        return f

    def get_tag_disk(self, tag):
        """
        Read a tag directly from disk. Can be slow, use with caution.

        Intended for use with large fields like covers and
        lyrics that shouldn't be loaded to the in-mem db.

        .. warning:: The track instance will not be updated with the new tag

        :returns: None if the tag does not exist
        """
        f = self._get_format_obj()
        if f:
            try:
                return f.read_tags([tag])[tag]
            except KeyError:
                return None

    def set_tag_disk(self, tag, values):
        """
        Set a tag directly to disk. Can be slow, use with caution.

        Intended for use with large fields like covers and
        lyrics that shouldn't be loaded to the in-mem db.

        .. warning:: The track instance will not be updated with the new tag

        :returns: None if the tag does not exist
        """
        f = self._get_format_obj()
        if f:
            values = self._xform_set_values(tag, values)
            f.write_tags({tag: values})

    def list_tags_disk(self):
        """
        List all the tags directly from file metadata. Can be slow,
        use with caution.
        """
        f = self._get_format_obj()
        if f:
            return f.get_keys_disk()

    ### convenience funcs for rating ###
    # these dont fit in the normal set of tag access methods,
    # but are sufficiently useful to warrant inclusion here

    def get_rating(self):
        """
        Returns the current track rating as an integer, as
        determined by the ``rating/maximum`` setting.

        :rtype: int
        """
        try:
            rating = float(self.get_tag_raw('__rating'))
        except (TypeError, KeyError, ValueError):
            return 0

        maximum = settings.get_option("rating/maximum", 5)
        rating = int(round(rating * float(maximum) / 100.0))

        if rating > maximum:
            return int(maximum)
        elif rating < 0:
            return 0

        return rating

    def set_rating(self, rating):
        """
        Sets the current track rating from an integer, on the
        scale determined by the ``rating/maximum`` setting.

        Returns the scaled rating
        """
        maximum = settings.get_option("rating/maximum", 5)
        rating = min(rating, maximum)
        rating = max(0, rating)
        rating = 100 * rating / maximum
        self.set_tags(__rating=rating)

        if self._write_rating_to_disk():
            self.set_tag_disk('__rating', rating)
        return rating

    ### Special functions for wrangling tag values ###

    @classmethod
    def format_sort(cls, values):
        if isinstance(values, list):
            return [cls.format_sort(v) for v in values]
        # order of these is important, both for speed and behavior!
        values = cls.strip_leading(values)
        values = cls.strip_marks(values)
        values = cls.lower(values)
        values = cls.the_cutter(values)
        values = cls.expand_doubles(values)
        return values

    @staticmethod
    def join_values(values, glue=" / "):
        """
        Exaile's standard method to join tag values
        """
        if isinstance(values, str):
            return values
        if isinstance(values, bytes):
            logger.warning(
                "join_values with bytes object is deprecated", stack_info=True
            )
            return xl.unicode.to_unicode(values)
        return glue.join(map(xl.unicode.to_unicode, values))

    @staticmethod
    def split_numerical(values):
        """
        this is used to split a tag like tracknumber that is in int/int
        format into its separate parts.

        input should be a string of the content, and may also
        be wrapped in a list.
        """
        if not values:
            return None, 0
        if isinstance(values, list):
            val = values[0]
        else:
            val = values
        if isinstance(val, int):
            return val, 0
        split = val.split("/")[:2]
        try:
            one = int(split[0])
        except ValueError:
            one = None
        try:
            two = int(split[1])
        except (IndexError, ValueError):
            two = 0
        return (one, two)

    @staticmethod
    def strip_leading(value):
        """
        Strip special chars off the beginning of a field. If
        stripping the chars leaves nothing the original field is
        returned with only whitespace removed.
        """
        stripped = xl.unicode.to_unicode(value).lstrip(
            " `~!@#$%^&*()_+-={}|[]\\\";'<>?,./"
        )
        if stripped:
            return stripped
        else:
            return value.lstrip()

    @staticmethod
    def the_cutter(value):
        """
        Cut common words like 'the' from the beginning of a tag so that
        they sort properly.
        """
        lowered = value.lower()
        for word in Track.__the_cuts:
            if not word.endswith("'"):
                word += ' '
            if lowered.startswith(word):
                value = value[len(word) :]
                break
        return value

    @staticmethod
    def strip_marks(value):
        """
        Remove accents, diacritics, etc.
        """
        # value is appended afterwards so that like-accented values
        # will sort together.
        normalize = unicodedata.normalize
        category = unicodedata.category
        return (
            ''.join([c for c in normalize('NFD', value) if category(c) != 'Mn'])
            + " "
            + value
        )

    @staticmethod
    def expand_doubles(value):
        """
        turns characters like æ into values suitable for sorting,
        like 'ae'. see _sortcharmap for the mapping.

        value must be a unicode object or this won't replace anything.

        value must be in lower-case
        """
        for k, v in _sortcharmap.items():
            value = value.replace(k, v)
        return value

    # This is slower, don't use it!
    #        return ''.join((_sortcharmap.get(c, c) for c in value))

    @staticmethod
    def lower(value):
        """
        Make tag value lower-case.
        """
        # add the original string after the lowered val so that
        # we sort case-sensitively if the case-insensitive value
        # are identical.
        return value.lower() + " " + value

    @classmethod
    def quoter(cls, value):
        """
        Escape quotes so that it's safe to put the value inside a
        "" literal for searches.

        Has no effect on non-string values.
        """
        try:
            return value.replace('"', '\\\"')
        except AttributeError:
            return value

    @classmethod
    def _the_cuts_cb(cls, name, obj, data):
        """
        PRIVATE

        update the cached the_cutter values
        """
        if data == "collection/strip_list":
            cls._Track__the_cuts = settings.get_option('collection/strip_list', [])

    ### Utility method intended for TrackDB ###

    @classmethod
    def _get_track_count(cls):
        '''Internal API, returns number of track objects we have'''
        return len(cls._Track__tracksdict)

    def _write_rating_to_disk(self):
        if not settings.get_option(
            'collection/write_rating_to_audio_file_metadata', False
        ):
            return False

        f = self._get_format_obj()
        if f is None:
            return False

        keys = f.tag_mapping
        if not '__rating' in keys:
            return False

        return True


event.add_callback(Track._the_cuts_cb, 'collection_option_set')
